# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

import datetime


def test_freezer(freezer):
    with freezer:
        assert datetime.datetime.now() == freezer.time_to_freeze



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

import warnings
from functools import wraps


def deprecated(message):
    def deprecated_decorator(func):
        @wraps(func)
        def deprecated_func(*args, **kwargs):
            warnings.warn(
                "'{0}' is a deprecated function. {1}".format(func.__name__, message),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter('default', DeprecationWarning)

            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from hashlib import md5
import json
import os
import random
import re
import subprocess
import time
import tempfile
import sys

import pytest
import requests

from datadog.util.compat import is_p3k, ConfigParser
from ..api.constants import API_KEY, APP_KEY, MONITOR_REFERENCED_IN_SLO_MESSAGE

WAIT_TIME = 11


def get_temp_file():
    """Return a (fn, fp) pair"""
    if is_p3k():
        fn = "/tmp/{0}-{1}".format(time.time(), random.random())
        return (fn, open(fn, "w+"))
    else:
        tf = tempfile.NamedTemporaryFile()
        return (tf.name, tf)


@pytest.fixture  # (scope="module")
def dogshell_config():
    config = ConfigParser()
    config.add_section("Connection")
    config.set("Connection", "apikey", API_KEY)
    config.set("Connection", "appkey", APP_KEY)
    config.set("Connection", "api_host", os.environ.get("DATADOG_HOST", "https://api.datadoghq.com"))
    return config


@pytest.fixture  # (scope="module")
def config_file(tmp_path, dogshell_config):
    """Generate a config file for the dog shell."""
    filename = tmp_path / ".test.dog.ini"
    with filename.open("w") as fp:
        dogshell_config.write(fp)
    return str(filename)


@pytest.fixture
def dogshell(capsys, config_file, dog):
    """Helper function to call the dog shell command."""
    import click
    from click.testing import CliRunner

    runner = CliRunner(mix_stderr=False)

    @click.command(context_settings={"ignore_unknown_options": True})
    @click.argument('args', nargs=-1, type=click.UNPROCESSED)
    def main(args):
        from datadog.dogshell import main as run
        orig = sys.argv
        try:
            sys.argv = list(args)
            run()
        finally:
            sys.argv = orig

    def run(args, stdin=None, check_return_code=True, use_cl_args=False):
        cmd = ["dogshell", "--config", config_file] + args
        if use_cl_args:
            cmd = [
                "dogshell",
                "--api-key={0}".format(dog._api_key),
                "--application-key={0}".format(dog._application_key),
            ] + args

        with capsys.disabled():
            result = runner.invoke(main, cmd, input=stdin, prog_name=cmd[0])
        return_code = result.exit_code
        out = result.stdout_bytes
        err = result.stderr_bytes
        if check_return_code:
            assert return_code == 0, err
            assert err == b""
        return out.decode("utf-8"), err.decode("utf-8"), return_code
    return run


@pytest.fixture
def dogshell_with_retry(vcr_cassette, dogshell):
    def run(cmd, retry_limit=10, retry_condition=lambda o, r: r != 0):
        number_of_interactions = len(vcr_cassette.data) if vcr_cassette.record_mode == "all" else -1

        out, err, return_code = dogshell(cmd, check_return_code=False)
        retry_count = 0
        while retry_count < retry_limit and retry_condition(out, return_code):
            time.sleep(WAIT_TIME)

            if vcr_cassette.record_mode == "all":
                # remove failed interactions
                vcr_cassette.data = vcr_cassette.data[:number_of_interactions]

            out, err, return_code = dogshell(cmd, check_return_code=False)
            retry_count += 1
        if retry_condition(out, return_code):
            raise Exception(
                "Retry limit reached for command {}:\nSTDOUT: {}\nSTDERR: {}\nSTATUS_CODE: {}".format(
                    cmd, out, err, return_code
                )
            )
        return out, err, return_code
    return run


@pytest.fixture
def get_unique(freezer, vcr_cassette_name, vcr_cassette, vcr):
    if vcr_cassette.record_mode == "all":
        seed = int(random.random() * 100000)

        with open(
            os.path.join(
                vcr.cassette_library_dir, vcr_cassette_name + ".seed"
            ),
            "w",
        ) as f:
            f.write(str(seed))
    else:
        with open(
            os.path.join(
                vcr.cassette_library_dir, vcr_cassette_name + ".seed"
            ),
            "r",
        ) as f:
            seed = int(f.readline().strip())

    random.seed(seed)

    def generate():
        with freezer:
            return md5(str(time.time() + random.random()).encode("utf-8")).hexdigest()
    return generate


class TestDogshell:
    host_name = "test.host.dogshell5"

    # Tests
    def test_config_args(self, dogshell):
        out, err, return_code = dogshell(["--help"], use_cl_args=True)
        assert 0 == return_code

    def test_comment(self, dogshell, dogshell_with_retry, user_handle):
        # Post a new comment
        cmd = ["comment", "post", user_handle]
        comment_msg = "yo dudes"
        post_data = {}
        out, _, _ = dogshell(cmd, stdin=comment_msg)
        post_data = self.parse_response(out)
        assert "id" in post_data
        assert "url" in post_data
        assert comment_msg in post_data["message"]

        # Read that comment from its id
        cmd = ["comment", "show", post_data["id"]]
        out, _, _ = dogshell_with_retry(cmd)
        show_data = self.parse_response(out)
        assert comment_msg in show_data["message"]

        # Update the comment
        cmd = ["comment", "update", post_data["id"], user_handle]
        new_comment = "nothing much"
        out, _, _ = dogshell(cmd, stdin=new_comment)
        update_data = self.parse_response(out)
        assert update_data["id"] == post_data["id"]
        assert new_comment in update_data["message"]

    def test_event(self, dog, dogshell, dogshell_with_retry):
        # Post an event
        title = "Testing events from dogshell"
        body = "%%%\n*Cool!*\n%%%\n"
        tags = "tag:a,tag:b"
        cmd = ["event", "post", title, "--tags", tags]
        event_id = None

        def match_permalink(out):
            match = re.match(r".*/event/event\?id=([0-9]*)", out, re.DOTALL) or re.match(
                r".*/event/jump_to\?event_id=([0-9]*)", out, re.DOTALL
            )
            if match:
                return match.group(1)
            else:
                return None

        out, err, return_code = dogshell(cmd, stdin=body)

        event_id = match_permalink(out)
        assert event_id

        # Retrieve the event
        cmd = ["event", "show", event_id]
        out, _, _ = dogshell_with_retry(cmd)
        event_id2 = match_permalink(out)
        assert event_id == event_id2

        # Get a real time from the event
        event = dog.Event.get(event_id)
        start = event["event"]["date_happened"] - 30 * 60
        end = event["event"]["date_happened"] + 1

        # Get a stream of events
        cmd = ["event", "stream", str(start), str(end), "--tags", tags]
        out, err, return_code = dogshell(cmd)
        event_ids = (match_permalink(l) for l in out.split("\n"))
        event_ids = set([e for e in event_ids if e])
        assert event_id in event_ids

    def test_metrics(self, dogshell, get_unique, dogshell_with_retry):
        # Submit a unique metric from a unique host
        unique = get_unique()
        metric = "test.dogshell.test_metric_{}".format(unique)
        host = "{}{}".format(self.host_name, unique)
        dogshell(["metric", "post", "--host", host, metric, "1"])

        # Query for the host and metric
        dogshell_with_retry(
            ["search", "query", unique], retry_condition=lambda o, r: host not in o or metric not in o
        )

        # Give the host some tags
        # The host tag association can take some time, so bump the retry limit to reduce flakiness
        tags0 = ["t0", "t1"]
        out, _, _ = dogshell_with_retry(["tag", "add", host] + tags0, retry_limit=30)
        for t in tags0:
            assert t in out

        # Verify that that host got those tags
        dogshell_with_retry(["tag", "show", host], retry_condition=lambda o, r: "t0" not in o or "t1" not in o)

        # Replace the tags with a different set
        tags1 = ["t2", "t3"]
        out, _, _ = dogshell(["tag", "replace", host] + tags1)
        for t in tags1:
            assert t in out
        for t in tags0:
            assert t not in out

        # Remove all the tags
        out, _, _ = dogshell(["tag", "detach", host])
        assert out == ""

    def test_timeboards(self, dogshell, get_unique):
        # Create a timeboard and write it to a file
        name, temp0 = get_temp_file()
        graph = {
            "title": "test metric graph",
            "definition": {"requests": [{"q": "testing.metric.1{host:blah.host.1}"}], "viz": "timeseries"},
        }
        dogshell(["timeboard", "new_file", name, json.dumps(graph)])
        dash = json.load(temp0)
        assert "id" in dash
        assert "title" in dash

        # Update the file and push it to the server
        unique = get_unique()
        dash["title"] = "dash title {}".format(unique)
        name, temp1 = get_temp_file()
        json.dump(dash, temp1)
        temp1.flush()
        dogshell(["timeboard", "push", temp1.name])

        # Query the server to verify the change
        out, _, _ = dogshell(["timeboard", "show", str(dash["id"])])

        out = json.loads(out)
        out["dash"]["id"] == dash["id"]
        out["dash"]["title"] == dash["title"]

        new_title = "new_title"
        new_desc = "new_desc"
        new_dash = [
            {
                "title": "blerg",
                "definition": {"viz": "timeseries", "requests": [{"q": "avg:system.load.15{web,env:prod}"}]},
            }
        ]

        # Update a dash directly on the server
        out, _, _ = dogshell(
            ["timeboard", "update", str(dash["id"]), new_title, new_desc], stdin=json.dumps(new_dash)
        )
        out = json.loads(out)
        # Template variables are empty, lets remove them because the `pull` command won't show them
        out["dash"].pop("template_variables", None)
        assert out["dash"]["id"] == dash["id"]
        assert out["dash"]["title"] == new_title
        assert out["dash"]["description"] == new_desc
        assert out["dash"]["graphs"] == new_dash

        # Pull the updated dash to disk
        fd, updated_file = tempfile.mkstemp()
        try:
            dogshell(["timeboard", "pull", str(dash["id"]), updated_file])
            updated_dash = {}
            with open(updated_file) as f:
                updated_dash = json.load(f)
            assert out["dash"] == updated_dash
        finally:
            os.unlink(updated_file)

        # Delete the dash
        dogshell(["--timeout", "30", "timeboard", "delete", str(dash["id"])])

        # Verify that it's not on the server anymore
        out, _, return_code = dogshell(["dashboard", "show", str(dash["id"])], check_return_code=False)
        assert return_code != 0

    def test_screenboards(self, dogshell, get_unique):
        # Create a screenboard and write it to a file
        name, temp0 = get_temp_file()
        graph = {
            "title": "test metric graph",
            "definition": {"requests": [{"q": "testing.metric.1{host:blah.host.1}"}], "viz": "timeseries"},
        }
        dogshell(["screenboard", "new_file", name, json.dumps(graph)])
        screenboard = json.load(temp0)

        assert "id" in screenboard
        assert "board_title" in screenboard

        # Update the file and push it to the server
        unique = get_unique()
        screenboard["title"] = "screenboard title {}".format(unique)
        name, temp1 = get_temp_file()
        json.dump(screenboard, temp1)
        temp1.flush()
        dogshell(["screenboard", "push", temp1.name])

        # Query the server to verify the change
        out, _, _ = dogshell(["screenboard", "show", str(screenboard["id"])])

        out = json.loads(out)
        assert out["id"] == screenboard["id"]
        assert out["title"] == screenboard["title"]

        new_title = "new_title"
        new_desc = "new_desc"
        new_screen = [{"title": "blerg", "definition": {"requests": [{"q": "avg:system.load.15{web,env:prod}"}]}}]

        # Update a screenboard directly on the server
        dogshell(
            ["screenboard", "update", str(screenboard["id"]), new_title, new_desc], stdin=json.dumps(new_screen)
        )
        # Query the server to verify the change
        out, _, _ = dogshell(["screenboard", "show", str(screenboard["id"])])
        out = json.loads(out)
        assert out["id"] == screenboard["id"]
        assert out["board_title"] == new_title
        assert out["description"] == new_desc
        assert out["widgets"] == new_screen

        # Pull the updated screenboard to disk
        fd, updated_file = tempfile.mkstemp()
        try:
            dogshell(["screenboard", "pull", str(screenboard["id"]), updated_file])
            updated_screenboard = {}
            with open(updated_file) as f:
                updated_screenboard = json.load(f)
            assert out == updated_screenboard
        finally:
            os.unlink(updated_file)

        # Share the screenboard
        out, _, _ = dogshell(["screenboard", "share", str(screenboard["id"])])
        out = json.loads(out)
        assert out["board_id"] == screenboard["id"]
        # Verify it's actually shared
        public_url = out["public_url"]
        response = requests.get(public_url)
        assert response.status_code == 200

        # Revoke the screenboard and verify it's actually revoked
        dogshell(["screenboard", "revoke", str(screenboard["id"])])
        response = requests.get(public_url)
        assert response.status_code == 404

        # Delete the screenboard
        dogshell(["--timeout", "30", "screenboard", "delete", str(screenboard["id"])])

        # Verify that it's not on the server anymore
        _, _, return_code = dogshell(["screenboard", "show", str(screenboard["id"])], check_return_code=False)
        assert return_code != 0

    # Test monitors
    @pytest.mark.admin_needed
    def test_monitors(self, dogshell):
        # Create a monitor
        query = "avg(last_1h):sum:system.net.bytes_rcvd{*} by {host} > 100"
        type_alert = "metric alert"
        out, _, _ = dogshell(["monitor", "post", type_alert, query])

        out = json.loads(out)
        assert out["query"] == query
        assert out["type"] == type_alert
        monitor_id = str(out["id"])
        monitor_name = out["name"]

        out, _, _ = dogshell(["monitor", "show", monitor_id])
        out = json.loads(out)
        assert out["query"] == query
        assert out["options"]["notify_no_data"] is False

        # Update options
        options = {"notify_no_data": True, "no_data_timeframe": 20}
        out, err, return_code = dogshell(
            ["monitor", "update", monitor_id, type_alert, query, "--options", json.dumps(options)],
            check_return_code=False
        )

        out = json.loads(out)
        assert query in out["query"]
        assert out["options"]["notify_no_data"] == options["notify_no_data"]
        assert out["options"]["no_data_timeframe"] == options["no_data_timeframe"]
        assert 'DEPRECATION' in err
        assert return_code == 0

        # Update message only
        updated_message = "monitor updated"
        current_options = out["options"]
        out, err, return_code = dogshell(
            ["monitor", "update", monitor_id, "--message", updated_message]
        )

        out = json.loads(out)
        assert updated_message == out["message"]
        assert query == out["query"]
        assert monitor_name == out["name"]
        assert current_options == out["options"]

        # Updating optional type and query
        updated_query = "avg(last_15m):sum:system.net.bytes_rcvd{*} by {env} > 222"
        updated_type = "query alert"

        out, err, return_code = dogshell(
            ["monitor", "update", monitor_id, "--type", updated_type, "--query", updated_query]
        )

        out = json.loads(out)
        assert updated_query in out["query"]
        assert updated_type in out["type"]
        assert updated_message in out["message"] # updated_message updated in previous step
        assert monitor_name in out["name"]
        assert current_options == out["options"]

        # Mute monitor
        out, _, _ = dogshell(["monitor", "mute", str(out["id"])])
        out = json.loads(out)
        assert str(out["id"]) == monitor_id
        assert out["options"]["silenced"] == {"*": None}

        # Unmute monitor
        out, _, _ = dogshell(["monitor", "unmute", "--all_scopes", monitor_id], check_return_code=False)
        out = json.loads(out)
        assert str(out["id"]) == monitor_id
        assert out["options"]["silenced"] == {}

        # Unmute all scopes of a monitor
        options = {"silenced": {"host:abcd1234": None, "host:abcd1235": None}}

        out, err, return_code = dogshell(
            ["monitor", "update", monitor_id, type_alert, query, "--options", json.dumps(options)],
            check_return_code=False
        )

        out = json.loads(out)
        assert out["query"] == query
        assert out["options"]["silenced"] == {"host:abcd1234": None, "host:abcd1235": None}
        assert "DEPRECATION" in err
        assert return_code == 0

        out, _, _ = dogshell(["monitor", "unmute", str(out["id"]), "--all_scopes"])
        out = json.loads(out)
        assert str(out["id"]) == monitor_id
        assert out["options"]["silenced"] == {}

        # Test can_delete monitor
        monitor_ids = [int(monitor_id)]
        str_monitor_ids = str(monitor_id)
        out, _, _ = dogshell(["monitor", "can_delete", str_monitor_ids])
        out = json.loads(out)
        assert out["data"]["ok"] == monitor_ids
        assert out["errors"] is None

        # Create a monitor-based SLO
        out, _, _ = dogshell(
            [
                "service_level_objective",
                "create",
                "--type",
                "monitor",
                "--monitor_ids",
                str_monitor_ids,
                "--name",
                "test_slo",
                "--thresholds",
                "7d:90",
            ]
        )
        out = json.loads(out)
        slo_id = out["data"][0]["id"]

        # Test can_delete monitor
        out, _, _ = dogshell(["monitor", "can_delete", str_monitor_ids])
        out = json.loads(out)
        assert out["data"]["ok"] == []
        # TODO update the error message template
        # assert out["errors"] == {
        #     str(monitor_id): [MONITOR_REFERENCED_IN_SLO_MESSAGE.format(monitor_id, slo_id)]
        # }

        # Delete a service_level_objective
        _, _, _ = dogshell(["service_level_objective", "delete", slo_id])

        # Test can_delete monitor
        out, _, _ = dogshell(["monitor", "can_delete", str_monitor_ids])
        out = json.loads(out)
        assert out["data"]["ok"] == monitor_ids
        assert out["errors"] is None

        # Delete a monitor
        dogshell(["monitor", "delete", monitor_id])
        # Verify that it's not on the server anymore
        _, _, return_code = dogshell(["monitor", "show", monitor_id], check_return_code=False)
        assert return_code != 0

        # Mute all
        out, _, _ = dogshell(["monitor", "mute_all"])
        out = json.loads(out)
        assert out["active"] is True

        # Unmute all
        dogshell(["monitor", "unmute_all"])
        # Retry unmuting all -> should raise an error this time
        _, _, return_code = dogshell(["monitor", "unmute_all"], check_return_code=False)
        assert return_code != 0

        # Test validate monitor
        monitor_type = "metric alert"
        valid_options = '{"thresholds": {"critical": 200.0}}'
        invalid_options = '{"thresholds": {"critical": 90.0}}'

        # Check with an invalid query.
        invalid_query = "THIS IS A BAD QUERY"
        out, _, _ = dogshell(["monitor", "validate", monitor_type, invalid_query, "--options", valid_options])
        out = json.loads(out)
        assert out == {"errors": ["The value provided for parameter 'query' is invalid"]}

        # Check with a valid query, invalid options.
        valid_query = "avg(last_1h):sum:system.net.bytes_rcvd{host:host0} > 200"
        out, _, _ = dogshell(["monitor", "validate", monitor_type, valid_query, "--options", invalid_options])
        out = json.loads(out)
        assert out == {"errors": ["Alert threshold (90.0) does not match that used in the query (200.0)."]}

        # Check with a valid query, valid options.
        out, _, _ = dogshell(["monitor", "validate", monitor_type, valid_query, "--options", valid_options])
        out = json.loads(out)
        assert out == {}

    def test_host_muting(self, freezer, dogshell, get_unique, dogshell_with_retry):
        # Submit a metric to create a host
        hostname = "my.test.host{}".format(get_unique())
        dogshell(["metric", "post", "--host", hostname, "metric", "1"])

        # Wait for the host to appear
        dogshell_with_retry(["tag", "show", hostname])

        message = "Muting this host for a test."
        with freezer:
            end = int(time.time()) + 60 * 60

        # Mute a host
        out, _, _ = dogshell(["host", "mute", hostname, "--message", message, "--end", str(end)])
        out = json.loads(out)
        assert out["action"] == "Muted"
        assert out["hostname"] == hostname
        assert out["message"] == message
        assert out["end"] == end

        # We shouldn't be able to mute a host that's already muted, unless we include
        # the override param.
        end2 = end + 60 * 15

        _, _, return_code = dogshell_with_retry(
            ["host", "mute", hostname, "--end", str(end2)], retry_condition=lambda o, r: r == 0
        )
        assert return_code != 0

        out, _, _ = dogshell(["host", "mute", hostname, "--end", str(end2), "--override"])
        out = json.loads(out)
        assert out["action"] == "Muted"
        assert out["hostname"] == hostname
        assert out["end"] == end2

        # Unmute a host
        out, _, _ = dogshell(["host", "unmute", hostname])
        out = json.loads(out)
        assert out["action"] == "Unmuted"
        assert out["hostname"] == hostname

    def test_downtime_schedule(self, freezer, dogshell):
        # Schedule a downtime
        scope = "env:staging"
        with freezer:
            start = str(int(time.time()))
        out, _, _ = dogshell(["downtime", "post", scope, start])
        out = json.loads(out)
        assert out["scope"][0] == scope
        assert out["disabled"] is False
        downtime_id = str(out["id"])

        # Get downtime
        out, _, _ = dogshell(["downtime", "show", downtime_id])
        out = json.loads(out)
        assert out["scope"][0] == scope
        assert out["disabled"] is False

        # Update downtime
        message = "Doing some testing on staging."
        with freezer:
            end = int(time.time()) + 60000
        out, _, _ = dogshell(
            ["downtime", "update", downtime_id, "--scope", scope, "--end", str(end), "--message", message]
        )
        out = json.loads(out)
        assert out["end"] == end
        assert out["message"] == message
        assert out["disabled"] is False

        # Cancel downtime
        dogshell(["downtime", "delete", downtime_id])

        # Get downtime and check if it is cancelled
        out, _, _ = dogshell(["downtime", "show", downtime_id])
        out = json.loads(out)
        assert out["scope"][0] == scope
        assert out["disabled"] is True

    def test_downtime_cancel_by_scope(self, dogshell):
        # Schedule a downtime
        scope = "env:staging"
        out, _, _ = dogshell(["downtime", "post", scope, str(int(time.time()))])
        out = json.loads(out)
        assert out["scope"][0] == scope
        assert out["disabled"] is False
        downtime_id = str(out["id"])

        # Cancel the downtime by scope
        dogshell(["downtime", "cancel_by_scope", scope])

        # Get downtime and check if it is cancelled
        out, _, _ = dogshell(["downtime", "show", downtime_id])
        out = json.loads(out)
        assert out["scope"][0] == scope
        assert out["disabled"] is True

    def test_service_check(self, dogshell):
        out, _, _ = dogshell(["service_check", "check", "check_pg", "host0", "1"])
        out = json.loads(out)
        assert out["status"], "ok"

    def parse_response(self, out):
        data = {}
        for line in out.split("\n"):
            parts = re.split(r"\s+", str(line).strip())
            key = parts[0]
            # Could potentially have errors with other whitespace
            val = " ".join(parts[1:])
            if key:
                data[key] = val
        return data



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
# stdlib
from io import BytesIO
import unittest
import json

# 3p
from mock import Mock
import requests

# datadog
from datadog import initialize, api
from datadog.api.http_client import RequestClient
from datadog.api.exceptions import ApiError
from datadog.api.resources import (
    CreateableAPIResource,
    UpdatableAPIResource,
    DeletableAPIResource,
    GetableAPIResource,
    ListableAPIResource,
    ListableAPISubResource,
    AddableAPISubResource,
    UpdatableAPISubResource,
    DeletableAPISubResource,
    ActionAPIResource
)
from datadog.util.compat import iteritems, is_p3k
from tests.util.contextmanagers import EnvVars


API_KEY = "apikey"
APP_KEY = "applicationkey"
API_HOST = "https://example.com"
HOST_NAME = "agent.hostname"
FAKE_PROXY = {
    "https": "http://user:pass@10.10.1.10:3128/",
}


class MockSession(object):
    """docstring for MockSession"""
    _args = None
    _kwargs = None
    _count = 0

    def request(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._count += 1
        return MockResponse()

    def call_args(self):
        return self._args, self._kwargs

    def call_count(self):
        return self._count


class MockResponse(requests.Response):

    def __init__(self, raise_for_status=False):
        super(MockResponse, self).__init__()
        self._raise_for_status = raise_for_status

    def raise_for_status(self):
        if not self._raise_for_status:
            return
        raise ApiError({'errors': ""})


# A few API Resources
class MyCreatable(CreateableAPIResource):
    _resource_name = 'creatables'

class MyParamsApiKeyCreatable(CreateableAPIResource):
    _resource_name = 'series'

class MyUpdatable(UpdatableAPIResource):
    _resource_name = 'updatables'


class MyGetable(GetableAPIResource):
    _resource_name = 'getables'


class MyListable(ListableAPIResource):
    _resource_name = 'listables'


class MyDeletable(DeletableAPIResource):
    _resource_name = 'deletables'


class MyListableSubResource(ListableAPISubResource):
    _resource_name = 'resource_name'
    _sub_resource_name = 'sub_resource_name'


class MyAddableSubResource(AddableAPISubResource):
    _resource_name = 'resource_name'
    _sub_resource_name = 'sub_resource_name'


class MyUpdatableSubResource(UpdatableAPISubResource):
    _resource_name = 'resource_name'
    _sub_resource_name = 'sub_resource_name'


class MyDeletableSubResource(DeletableAPISubResource):
    _resource_name = 'resource_name'
    _sub_resource_name = 'sub_resource_name'


class MyActionable(ActionAPIResource):
    _resource_name = 'actionables'

    @classmethod
    def trigger_class_action(cls, method, name, id=None, params=None, **body):
        super(MyActionable, cls)._trigger_class_action(method, name, id, params, **body)

    @classmethod
    def trigger_action(cls, method, name, id=None, **body):
        super(MyActionable, cls)._trigger_action(method, name, id, **body)


# Test classes
class DatadogAPITestCase(unittest.TestCase):

    def setUp(self):
        # Mock patch requests
        self.request_mock = MockSession()
        RequestClient._session = self.request_mock
        # self.request_patcher = patch('requests.Session')
        # request_class_mock = self.request_patcher.start()
        # self.request_mock = request_class_mock.return_value
        # self.request_mock.request = Mock(return_value=MockResponse())

    def tearDown(self):
        RequestClient._session = None

    def load_request_response(self, status_code=200, response_body='{}', raise_for_status=False):
        """
        Load the repsonse body from the given payload
        """
        mock_response = MockResponse(raise_for_status=raise_for_status)
        if is_p3k():
            mock_response.raw = BytesIO(bytes(response_body, 'utf-8'))
        else:
            mock_response.raw = BytesIO(response_body)
        mock_response.status_code = status_code

        self.request_mock.request = Mock(return_value=mock_response)

    def arm_requests_to_raise(self):
        """
        Arm the mocked request to raise for status.
        """
        self.request_mock.request = Mock(return_value=MockResponse(raise_for_status=True))

    def get_request_data(self):
        """
        Returns JSON formatted data from the submitted `requests`.
        """
        _, kwargs = self.request_mock.call_args()
        return json.loads(kwargs['data'])

    def request_called_with(self, method, url, data=None, params=None):
        (req_method, req_url), others = self.request_mock.call_args()
        self.assertEqual(method, req_method, req_method)
        self.assertEqual(url, req_url, req_url)

        if data:
            self.assertIn('data', others)
            others_data = json.loads(others['data'])
            self.assertEqual(data, others_data, others['data'])

        if params:
            self.assertIn('params', others)
            for (k, v) in iteritems(params):
                self.assertIn(k, others['params'], others['params'])
                self.assertEqual(v, others['params'][k])

    def assertIn(self, first, second, msg=None):
        msg = msg or "{0} not in {1}".format(first, second)
        self.assertTrue(first in second, msg)


class DatadogAPINoInitialization(DatadogAPITestCase):
    def tearDown(self):
        super(DatadogAPINoInitialization, self).tearDown()
        # Restore default values
        api._api_key = None
        api._application_key = None
        api._api_host = None
        api._host_name = None
        api._proxies = None

    def setUp(self):
        super(DatadogAPINoInitialization, self).setUp()
        api._api_key = api._application_key = api._host_name = api._api_host = None


class DatadogAPIWithInitialization(DatadogAPITestCase):
    def setUp(self):
        super(DatadogAPIWithInitialization, self).setUp()
        initialize(api_key=API_KEY, app_key=APP_KEY, api_host=API_HOST)

    def tearDown(self):
        super(DatadogAPIWithInitialization, self).tearDown()
        # Restore default values
        api._api_key = None
        api._application_key = None
        api._api_host = None
        api._host_name = None
        api._proxies = None



# -*- coding: utf-8 -*-
# pylint: disable=line-too-long,too-many-public-methods

# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
Tests for dogstatsd.py
"""
# Standard libraries
from collections import deque
from contextlib import closing
from threading import Thread
import errno
import os
import shutil
import socket
import tempfile
import threading
import time
import unittest
import warnings

# Third-party libraries
import mock
from mock import call, Mock, mock_open, patch
import pytest

# Datadog libraries
from datadog import initialize, statsd
from datadog import __version__ as version
from datadog.dogstatsd.base import DEFAULT_FLUSH_INTERVAL, DogStatsd, MIN_SEND_BUFFER_SIZE, UDP_OPTIMAL_PAYLOAD_LENGTH, UDS_OPTIMAL_PAYLOAD_LENGTH
from datadog.dogstatsd.context import TimedContextManagerDecorator
from datadog.util.compat import is_higher_py35, is_p3k
from tests.util.contextmanagers import preserve_environment_variable, EnvVars
from tests.unit.dogstatsd.fixtures import load_fixtures


class FakeSocket(object):
    """ A fake socket for testing. """

    FLUSH_GRACE_PERIOD = 0.2

    def __init__(self, flush_interval=DEFAULT_FLUSH_INTERVAL):
        self.payloads = deque()

        self._flush_interval = flush_interval
        self._flush_wait = False

    def send(self, payload):
        if is_p3k():
            assert isinstance(payload, bytes)
        else:
            assert isinstance(payload, str)

        self.payloads.append(payload)

    def recv(self, count=1, reset_wait=False, no_wait=False):
        # Initial receive should wait for the flush thread timeout unless we
        # specifically want either a follow-up wait or no waiting at all
        if not self._flush_wait or reset_wait:
            if not no_wait:
                time.sleep(self._flush_interval+self.FLUSH_GRACE_PERIOD)
            self._flush_wait = True

        if count > len(self.payloads):
            return None

        out = []
        for _ in range(count):
            out.append(self.payloads.popleft().decode('utf-8'))
        return '\n'.join(out)

    def close(self):
        pass

    def __repr__(self):
        return str(self.payloads)


class BrokenSocket(FakeSocket):
    def __init__(self, error_number=None):
        super(BrokenSocket, self).__init__()

        self.error_number = error_number

    def send(self, payload):
        error = socket.error("Socket error [Errno {}]".format(self.error_number))
        if self.error_number:
            error.errno = self.error_number

        raise error


class OverflownSocket(BrokenSocket):

    def __init__(self):
        super(OverflownSocket, self).__init__(errno.EAGAIN)


def telemetry_metrics(metrics=1, events=0, service_checks=0, bytes_sent=0, bytes_dropped=0, packets_sent=1, packets_dropped=0, transport="udp", tags=""):
    tags = "," + tags if tags else ""

    return "\n".join([
        "datadog.dogstatsd.client.metrics:{}|c|#client:py,client_version:{},client_transport:{}{}".format(metrics, version, transport, tags),
        "datadog.dogstatsd.client.events:{}|c|#client:py,client_version:{},client_transport:{}{}".format(events, version, transport, tags),
        "datadog.dogstatsd.client.service_checks:{}|c|#client:py,client_version:{},client_transport:{}{}".format(service_checks, version, transport, tags),
        "datadog.dogstatsd.client.bytes_sent:{}|c|#client:py,client_version:{},client_transport:{}{}".format(bytes_sent, version, transport, tags),
        "datadog.dogstatsd.client.bytes_dropped:{}|c|#client:py,client_version:{},client_transport:{}{}".format(bytes_dropped, version, transport, tags),
        "datadog.dogstatsd.client.packets_sent:{}|c|#client:py,client_version:{},client_transport:{}{}".format(packets_sent, version, transport, tags),
        "datadog.dogstatsd.client.packets_dropped:{}|c|#client:py,client_version:{},client_transport:{}{}".format(packets_dropped, version, transport, tags)
    ]) + "\n"


class TestDogStatsd(unittest.TestCase):
    METRIC_TYPE_MAP = {
        'gauge': { 'id': 'g' },
        'timing': { 'id': 'ms' },
    }

    def setUp(self):
        """
        Set up a default Dogstatsd instance and mock the proc filesystem.
        """
        #
        self.statsd = DogStatsd(telemetry_min_flush_interval=0)
        self.statsd.socket = FakeSocket()
        self.statsd._reset_telemetry()

        # Mock the proc filesystem
        route_data = load_fixtures('route')
        self._procfs_mock = patch('datadog.util.compat.builtins.open', mock_open())
        self._procfs_mock.start().return_value.readlines.return_value = route_data.split("\n")

    def tearDown(self):
        """
        Unmock the proc filesystem.
        """
        self._procfs_mock.stop()

    def assert_equal_telemetry(self, expected_payload, actual_payload, telemetry=None):
        if telemetry is None:
            telemetry = telemetry_metrics(bytes_sent=len(expected_payload))

        if expected_payload:
            expected_payload = "\n".join([expected_payload, telemetry])
        else:
            expected_payload = telemetry

        return self.assertEqual(expected_payload, actual_payload)

    def send_and_assert(
        self,
        dogstatsd,
        expected_metrics,
        last_telemetry_size=0,
        buffered=False,
    ):
        """
        Send and then asserts that a chain of metrics arrive in the right order
        and with expected telemetry values.
        """

        expected_messages = []
        for metric_type, metric_name, metric_value in expected_metrics:
            # Construct the expected message data
            metric_type_id = TestDogStatsd.METRIC_TYPE_MAP[metric_type]['id']
            expected_messages.append(
                "{}:{}|{}\n".format(metric_name, metric_value, metric_type_id)
            )

            # Send the value
            getattr(dogstatsd, metric_type)(metric_name, metric_value)

        # Sanity check
        if buffered:
            # Ensure that packets didn't arrive immediately if we are expecting
            # buffering behavior
            self.assertIsNone(dogstatsd.socket.recv(2, no_wait=True))

        metrics = 1
        if buffered:
            metrics = len(expected_messages)

        if buffered:
            expected_messages = [ ''.join(expected_messages) ]

        for message in expected_messages:
            packets_sent = 1
            # For all ono-initial packets, our current telemetry stats will
            # contain the metadata for the last telemetry packet as well.
            if last_telemetry_size > 0:
                packets_sent += 1

            expected_metrics=telemetry_metrics(
                metrics=metrics,
                packets_sent=packets_sent,
                bytes_sent=len(message) + last_telemetry_size
            )
            self.assert_equal_telemetry(
                message,
                dogstatsd.socket.recv(2, no_wait=not buffered, reset_wait=True),
                telemetry=expected_metrics,
            )
            last_telemetry_size = len(expected_metrics)

        return last_telemetry_size

    def assert_almost_equal(self, val1, val2, delta):
        """
        Calculates a delta between first and second value and ensures
        that this difference falls within the delta range
        """
        return self.assertTrue(
            0 <= abs(val1 - val2) <= delta,
            "Absolute difference of {} and {} ({}) is not within {}".format(
                val1,
                val2,
                abs(val1-val2),
                delta,
            ),
        )

    def recv(self, *args, **kwargs):
        return self.statsd.socket.recv(*args, **kwargs)

    def test_initialization(self):
        """
        `initialize` overrides `statsd` default instance attributes.
        """
        options = {
            'statsd_host': "myhost",
            'statsd_port': 1234
        }

        # Default values
        self.assertEqual(statsd.host, "localhost")
        self.assertEqual(statsd.port, 8125)

        # After initialization
        initialize(**options)
        self.assertEqual(statsd.host, "myhost")
        self.assertEqual(statsd.port, 1234)

        # Add namespace
        options['statsd_namespace'] = "mynamespace"
        initialize(**options)
        self.assertEqual(statsd.host, "myhost")
        self.assertEqual(statsd.port, 1234)
        self.assertEqual(statsd.namespace, "mynamespace")

        # Set `statsd` host to the system's default route
        initialize(statsd_use_default_route=True, **options)
        self.assertEqual(statsd.host, "172.17.0.1")
        self.assertEqual(statsd.port, 1234)

        # Add UNIX socket
        options['statsd_socket_path'] = '/var/run/dogstatsd.sock'
        initialize(**options)
        self.assertEqual(statsd.socket_path, options['statsd_socket_path'])
        self.assertIsNone(statsd.host)
        self.assertIsNone(statsd.port)

    def test_dogstatsd_initialization_with_env_vars(self):
        """
        Dogstatsd can retrieve its config from env vars when
        not provided in constructor.
        """
        # Setup
        with preserve_environment_variable('DD_AGENT_HOST'):
            os.environ['DD_AGENT_HOST'] = 'myenvvarhost'
            with preserve_environment_variable('DD_DOGSTATSD_PORT'):
                os.environ['DD_DOGSTATSD_PORT'] = '4321'
                dogstatsd = DogStatsd()

        # Assert
        self.assertEqual(dogstatsd.host, "myenvvarhost")
        self.assertEqual(dogstatsd.port, 4321)

    def test_default_route(self):
        """
        Dogstatsd host can be dynamically set to the default route.
        """
        self.assertEqual(
            DogStatsd(use_default_route=True).host,
            "172.17.0.1"
        )

    def test_set(self):
        self.statsd.set('set', 123)
        self.assert_equal_telemetry('set:123|s\n', self.recv(2))

    def test_gauge(self):
        self.statsd.gauge('gauge', 123.4)
        self.assert_equal_telemetry('gauge:123.4|g\n', self.recv(2))

    def test_counter(self):
        self.statsd.increment('page.views')
        self.statsd.flush()
        self.assert_equal_telemetry('page.views:1|c\n', self.recv(2))

        self.statsd._reset_telemetry()
        self.statsd.increment('page.views', 11)
        self.statsd.flush()
        self.assert_equal_telemetry('page.views:11|c\n', self.recv(2))

        self.statsd._reset_telemetry()
        self.statsd.decrement('page.views')
        self.statsd.flush()
        self.assert_equal_telemetry('page.views:-1|c\n', self.recv(2))

        self.statsd._reset_telemetry()
        self.statsd.decrement('page.views', 12)
        self.statsd.flush()
        self.assert_equal_telemetry('page.views:-12|c\n', self.recv(2))

    def test_histogram(self):
        self.statsd.histogram('histo', 123.4)
        self.assert_equal_telemetry('histo:123.4|h\n', self.recv(2))

    def test_pipe_in_tags(self):
        self.statsd.gauge('gt', 123.4, tags=['pipe|in:tag', 'red'])
        self.assert_equal_telemetry('gt:123.4|g|#pipe_in:tag,red\n', self.recv(2))

    def test_tagged_gauge(self):
        self.statsd.gauge('gt', 123.4, tags=['country:china', 'age:45', 'blue'])
        self.assert_equal_telemetry('gt:123.4|g|#country:china,age:45,blue\n', self.recv(2))

    def test_tagged_counter(self):
        self.statsd.increment('ct', tags=[u'country:españa', 'red'])
        self.assert_equal_telemetry(u'ct:1|c|#country:españa,red\n', self.recv(2))

    def test_tagged_histogram(self):
        self.statsd.histogram('h', 1, tags=['red'])
        self.assert_equal_telemetry('h:1|h|#red\n', self.recv(2))

    def test_sample_rate(self):
        # Disabling telemetry since sample_rate imply randomness
        self.statsd._telemetry = False

        self.statsd.increment('c', sample_rate=0)
        self.assertFalse(self.recv())

        for _ in range(10000):
            self.statsd.increment('sampled_counter', sample_rate=0.3)

        self.statsd.flush()

        total_metrics = 0
        payload = self.recv()
        while payload:
            metrics = payload.rstrip('\n').split('\n')
            for metric in metrics:
                self.assertEqual('sampled_counter:1|c|@0.3', metric)
            total_metrics += len(metrics)
            payload = self.recv()

        self.assert_almost_equal(3000, total_metrics, 150)

    def test_default_sample_rate(self):
        # Disabling telemetry since sample_rate imply randomness
        self.statsd._telemetry = False

        self.statsd.default_sample_rate = 0.3
        for _ in range(10000):
            self.statsd.increment('sampled_counter')

        total_metrics = 0
        payload = self.recv()
        while payload:
            metrics = payload.rstrip('\n').split('\n')
            for metric in metrics:
                self.assertEqual('sampled_counter:1|c|@0.3', metric)

            total_metrics += len(metrics)
            payload = self.recv()

        self.assert_almost_equal(3000, total_metrics, 150)

    def test_tags_and_samples(self):
        # Disabling telemetry since sample_rate imply randomness
        self.statsd._telemetry = False

        for _ in range(100):
            self.statsd.gauge('gst', 23, tags=["sampled"], sample_rate=0.9)

        self.assertEqual('gst:23|g|@0.9|#sampled', self.recv().split('\n')[0])

    def test_timing(self):
        self.statsd.timing('t', 123)
        self.assert_equal_telemetry('t:123|ms\n', self.recv(2))

    def test_event(self):
        self.statsd.event(
            'Title',
            u'L1\nL2',
            priority='low',
            date_happened=1375296969,
        )
        event2 = u'_e{5,6}:Title|L1\\nL2|d:1375296969|p:low\n'
        self.assert_equal_telemetry(
            event2,
            self.recv(2),
            telemetry=telemetry_metrics(
                metrics=0,
                events=1,
                bytes_sent=len(event2),
            ),
        )

        self.statsd._reset_telemetry()

        self.statsd.event('Title', u'♬ †øU †øU ¥ºu T0µ ♪',
                          aggregation_key='key', tags=['t1', 't2:v2'])
        event3 = u'_e{5,32}:Title|♬ †øU †øU ¥ºu T0µ ♪|k:key|#t1,t2:v2\n'
        self.assert_equal_telemetry(
            event3,
            self.recv(2, reset_wait=True),
            telemetry=telemetry_metrics(
                metrics=0,
                events=1,
                bytes_sent=len(event3),
            ),
        )

    def test_unicode_event(self):
        self.statsd.event(
                'my.prefix.Delivery - Daily Settlement Summary Report Delivery — Invoice Cloud succeeded',
                'Delivered — destination.csv')
        event = u'_e{89,29}:my.prefix.Delivery - Daily Settlement Summary Report Delivery — Invoice Cloud succeeded|' + \
            u'Delivered — destination.csv\n'
        self.assert_equal_telemetry(
            event,
            self.recv(2),
            telemetry=telemetry_metrics(
                metrics=0,
                events=1,
                bytes_sent=len(event),
            ),
        )

        self.statsd._reset_telemetry()

    def test_event_constant_tags(self):
        self.statsd.constant_tags = ['bar:baz', 'foo']
        self.statsd.event('Title', u'L1\nL2', priority='low', date_happened=1375296969)
        event = u'_e{5,6}:Title|L1\\nL2|d:1375296969|p:low|#bar:baz,foo\n'
        self.assert_equal_telemetry(
            event,
            self.recv(2),
            telemetry=telemetry_metrics(
                metrics=0,
                events=1,
                tags="bar:baz,foo",
                bytes_sent=len(event),
            ),
        )

        self.statsd._reset_telemetry()

        self.statsd.event('Title', u'♬ †øU †øU ¥ºu T0µ ♪',
                          aggregation_key='key', tags=['t1', 't2:v2'])
        event = u'_e{5,32}:Title|♬ †øU †øU ¥ºu T0µ ♪|k:key|#t1,t2:v2,bar:baz,foo\n'
        self.assert_equal_telemetry(
            event,
            self.recv(2, reset_wait=True),
            telemetry=telemetry_metrics(
                metrics=0,
                events=1,
                tags="bar:baz,foo",
                bytes_sent=len(event),
            ),
        )

    def test_service_check(self):
        now = int(time.time())
        self.statsd.service_check(
            'my_check.name', self.statsd.WARNING,
            tags=['key1:val1', 'key2:val2'], timestamp=now,
            hostname='i-abcd1234', message=u"♬ †øU \n†øU ¥ºu|m: T0µ ♪")
        check = u'_sc|my_check.name|{0}|d:{1}|h:i-abcd1234|#key1:val1,key2:val2|m:{2}'.format(self.statsd.WARNING, now, u'♬ †øU \\n†øU ¥ºu|m\\: T0µ ♪\n')
        self.assert_equal_telemetry(
            check,
            self.recv(2),
            telemetry=telemetry_metrics(
                metrics=0,
                service_checks=1,
                bytes_sent=len(check),
            ),
        )

    def test_service_check_constant_tags(self):
        self.statsd.constant_tags = ['bar:baz', 'foo']
        now = int(time.time())
        self.statsd.service_check(
            'my_check.name', self.statsd.WARNING,
            timestamp=now,
            hostname='i-abcd1234', message=u"♬ †øU \n†øU ¥ºu|m: T0µ ♪")
        check = u'_sc|my_check.name|{0}|d:{1}|h:i-abcd1234|#bar:baz,foo|m:{2}'.format(self.statsd.WARNING, now, u"♬ †øU \\n†øU ¥ºu|m\\: T0µ ♪\n")
        self.assert_equal_telemetry(
            check,
            self.recv(2, True),
            telemetry=telemetry_metrics(
                metrics=0,
                service_checks=1,
                tags="bar:baz,foo",
                bytes_sent=len(check),
            ),
        )

        self.statsd._reset_telemetry()

        self.statsd.service_check(
            'my_check.name', self.statsd.WARNING,
            tags=['key1:val1', 'key2:val2'], timestamp=now,
            hostname='i-abcd1234', message=u"♬ †øU \n†øU ¥ºu|m: T0µ ♪")
        check = u'_sc|my_check.name|{0}|d:{1}|h:i-abcd1234|#key1:val1,key2:val2,bar:baz,foo|m:{2}'.format(self.statsd.WARNING, now, u"♬ †øU \\n†øU ¥ºu|m\\: T0µ ♪\n")
        self.assert_equal_telemetry(
            check,
            self.recv(2, True),
            telemetry=telemetry_metrics(
                metrics=0,
                service_checks=1,
                tags="bar:baz,foo",
                bytes_sent=len(check),
            ),
        )

    def test_metric_namespace(self):
        """
        Namespace prefixes all metric names.
        """
        self.statsd.namespace = "foo"
        self.statsd.gauge('gauge', 123.4)
        self.assert_equal_telemetry('foo.gauge:123.4|g\n', self.recv(2))

    # Test Client level contant tags
    def test_gauge_constant_tags(self):
        self.statsd.constant_tags = ['bar:baz', 'foo']
        self.statsd.gauge('gauge', 123.4)
        metric = 'gauge:123.4|g|#bar:baz,foo\n'
        self.assert_equal_telemetry(metric, self.recv(2), telemetry=telemetry_metrics(tags="bar:baz,foo", bytes_sent=len(metric)))

    def test_counter_constant_tag_with_metric_level_tags(self):
        self.statsd.constant_tags = ['bar:baz', 'foo']
        self.statsd.increment('page.views', tags=['extra'])
        metric = 'page.views:1|c|#extra,bar:baz,foo\n'
        self.assert_equal_telemetry(metric, self.recv(2), telemetry=telemetry_metrics(tags="bar:baz,foo", bytes_sent=len(metric)))

    def test_gauge_constant_tags_with_metric_level_tags_twice(self):
        metric_level_tag = ['foo:bar']
        self.statsd.constant_tags = ['bar:baz']
        self.statsd.gauge('gauge', 123.4, tags=metric_level_tag)
        metric = 'gauge:123.4|g|#foo:bar,bar:baz\n'
        self.assert_equal_telemetry(
            metric,
            self.recv(2),
            telemetry=telemetry_metrics(
                tags="bar:baz",
                bytes_sent=len(metric),
            ),
        )

        self.statsd._reset_telemetry()

        # sending metrics multiple times with same metric-level tags
        # should not duplicate the tags being sent
        self.statsd.gauge('gauge', 123.4, tags=metric_level_tag)
        metric = 'gauge:123.4|g|#foo:bar,bar:baz\n'
        self.assert_equal_telemetry(
            metric,
            self.recv(2, reset_wait=True),
            telemetry=telemetry_metrics(
                tags="bar:baz",
                bytes_sent=len(metric),
            ),
        )

    def test_socket_error(self):
        self.statsd.socket = BrokenSocket()
        with mock.patch("datadog.dogstatsd.base.log") as mock_log:
            self.statsd.gauge('no error', 1)
            self.statsd.flush()

            mock_log.error.assert_not_called()
            mock_log.warning.assert_called_once_with(
                "Error submitting packet: %s, dropping the packet and closing the socket",
                mock.ANY,
            )

    def test_socket_overflown(self):
        self.statsd.socket = OverflownSocket()
        with mock.patch("datadog.dogstatsd.base.log") as mock_log:
            self.statsd.gauge('no error', 1)
            self.statsd.flush()

            mock_log.error.assert_not_called()
            calls = [call("Socket send would block: %s, dropping the packet", mock.ANY)]
            mock_log.debug.assert_has_calls(calls * 2)

    def test_socket_message_too_long(self):
        self.statsd.socket = BrokenSocket(error_number=errno.EMSGSIZE)
        with mock.patch("datadog.dogstatsd.base.log") as mock_log:
            self.statsd.gauge('no error', 1)
            self.statsd.flush()

            mock_log.error.assert_not_called()
            calls = [
                call(
                    "Packet size too big (size: %d): %s, dropping the packet",
                    mock.ANY,
                    mock.ANY,
                ),
            ]
            mock_log.debug.assert_has_calls(calls * 2)

    def test_socket_no_buffer_space(self):
        self.statsd.socket = BrokenSocket(error_number=errno.ENOBUFS)
        with mock.patch("datadog.dogstatsd.base.log") as mock_log:
            self.statsd.gauge('no error', 1)
            self.statsd.flush()

            mock_log.error.assert_not_called()
            calls = [call("Socket buffer full: %s, dropping the packet", mock.ANY)]
            mock_log.debug.assert_has_calls(calls * 2)

    @patch('socket.socket')
    def test_uds_socket_ensures_min_receive_buffer(self, mock_socket_create):
        mock_socket = mock_socket_create.return_value
        mock_socket.setblocking.return_value = None
        mock_socket.connect.return_value = None
        mock_socket.getsockopt.return_value = MIN_SEND_BUFFER_SIZE / 2

        datadog = DogStatsd(socket_path="/fake/uds/socket/path")
        datadog.gauge('some value', 1)
        datadog.flush()

        # Sanity check
        mock_socket_create.assert_called_once_with(socket.AF_UNIX, socket.SOCK_DGRAM)

        mock_socket.setsockopt.assert_called_once_with(
            socket.SOL_SOCKET,
            socket.SO_SNDBUF,
            MIN_SEND_BUFFER_SIZE,
        )

    @patch('socket.socket')
    def test_udp_socket_ensures_min_receive_buffer(self, mock_socket_create):
        mock_socket = mock_socket_create.return_value
        mock_socket.setblocking.return_value = None
        mock_socket.connect.return_value = None
        mock_socket.getsockopt.return_value = MIN_SEND_BUFFER_SIZE / 2

        datadog = DogStatsd()
        datadog.gauge('some value', 1)
        datadog.flush()

        # Sanity check
        mock_socket_create.assert_called_once_with(socket.AF_INET, socket.SOCK_DGRAM)

        mock_socket.setsockopt.assert_called_once_with(
            socket.SOL_SOCKET,
            socket.SO_SNDBUF,
            MIN_SEND_BUFFER_SIZE,
        )

    def test_distributed(self):
        """
        Measure the distribution of a function's run time using distribution custom metric.
        """
        # In seconds
        @self.statsd.distributed('distributed.test')
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.1)
            return (arg1, arg2, kwarg1, kwarg2)

        self.assertEqual('func', func.__name__)
        self.assertEqual('docstring', func.__doc__)

        result = func(1, 2, kwarg2=3)
        # Assert it handles args and kwargs correctly.
        self.assertEqual(result, (1, 2, 1, 3))

        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('d', type_)
        self.assertEqual('distributed.test', name)
        self.assert_almost_equal(0.1, float(value), 0.09)

        # Repeat, force timer value in milliseconds
        @self.statsd.distributed('distributed.test', use_ms=True)
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.5)
            return (arg1, arg2, kwarg1, kwarg2)

        func(1, 2, kwarg2=3)

        # Ignore telemetry packet
        packet = self.recv(2, reset_wait=True).split("\n")[0]
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('d', type_)
        self.assertEqual('distributed.test', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_timed(self):
        """
        Measure the distribution of a function's run time.
        """
        # In seconds
        @self.statsd.timed('timed.test')
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.5)
            return (arg1, arg2, kwarg1, kwarg2)

        self.assertEqual('func', func.__name__)
        self.assertEqual('docstring', func.__doc__)

        result = func(1, 2, kwarg2=3)
        # Assert it handles args and kwargs correctly.
        self.assertEqual(result, (1, 2, 1, 3))

        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(0.5, float(value), 0.1)

        # Repeat, force timer value in milliseconds
        @self.statsd.timed('timed.test', use_ms=True)
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.5)
            return (arg1, arg2, kwarg1, kwarg2)

        func(1, 2, kwarg2=3)
        self.statsd.flush()

        # Ignore telemetry packet
        packet = self.recv(2).split("\n")[0]
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_timed_in_ms(self):
        """
        Timed value is reported in ms when statsd.use_ms is True.
        """
        # Arm statsd to use_ms
        self.statsd.use_ms = True

        # Sample a function run time
        @self.statsd.timed('timed.test')
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.5)
            return (arg1, arg2, kwarg1, kwarg2)

        func(1, 2, kwarg2=3)

        # Assess the packet
        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(500, float(value), 100)

        # Repeat, force timer value in seconds
        @self.statsd.timed('timed.test', use_ms=False)
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.5)
            return (arg1, arg2, kwarg1, kwarg2)

        func(1, 2, kwarg2=3)
        self.statsd.flush()

        packet = self.recv()
        name_value, type_ = packet.rstrip('\n').split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(0.5, float(value), 0.1)

    def test_timed_no_metric(self, ):
        """
        Test using a decorator without providing a metric.
        """

        @self.statsd.timed()
        def func(arg1, arg2, kwarg1=1, kwarg2=1):
            """docstring"""
            time.sleep(0.5)
            return (arg1, arg2, kwarg1, kwarg2)

        self.assertEqual('func', func.__name__)
        self.assertEqual('docstring', func.__doc__)

        result = func(1, 2, kwarg2=3)
        # Assert it handles args and kwargs correctly.
        self.assertEqual(result, (1, 2, 1, 3))

        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('tests.unit.dogstatsd.test_statsd.func', name)
        self.assert_almost_equal(0.5, float(value), 0.1)

    @unittest.skipIf(not is_higher_py35(), reason="Coroutines are supported on Python 3.5 or higher.")
    def test_timed_coroutine(self):
        """
        Measure the distribution of a coroutine function's run time.

        Warning: Python > 3.5 only.
        """
        import asyncio

        source = """
@self.statsd.timed('timed.test')
async def print_foo():
    "docstring"
    import time
    time.sleep(0.5)
    print("foo")
        """
        exec(source, {}, locals())

        loop = asyncio.get_event_loop()
        loop.run_until_complete(locals()['print_foo']())
        loop.close()

        # Assert
        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed.test', name)
        self.assert_almost_equal(0.5, float(value), 0.1)

    def test_timed_context(self):
        """
        Measure the distribution of a context's run time.
        """
        # In seconds
        with self.statsd.timed('timed_context.test') as timer:
            self.assertTrue(isinstance(timer, TimedContextManagerDecorator))
            time.sleep(0.5)

        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test', name)
        self.assert_almost_equal(0.5, float(value), 0.1)
        self.assert_almost_equal(0.5, timer.elapsed, 0.1)

        # In milliseconds
        with self.statsd.timed('timed_context.test', use_ms=True) as timer:
            time.sleep(0.5)

        packet = self.recv(2, reset_wait=True).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test', name)
        self.assert_almost_equal(500, float(value), 100)
        self.assert_almost_equal(500, timer.elapsed, 100)

    def test_timed_context_exception(self):
        """
        Exception bubbles out of the `timed` context manager.
        """
        class ContextException(Exception):
            pass

        def func(self):
            with self.statsd.timed('timed_context.test.exception'):
                time.sleep(0.5)
                raise ContextException()

        # Ensure the exception was raised.
        with pytest.raises(ContextException):
            func(self)

        # Ensure the timing was recorded.
        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test.exception', name)
        self.assert_almost_equal(0.5, float(value), 0.1)

    def test_timed_context_no_metric_exception(self):
        """Test that an exception occurs if using a context manager without a metric."""

        def func(self):
            with self.statsd.timed():
                time.sleep(0.5)

        # Ensure the exception was raised.
        with pytest.raises(TypeError):
            func(self)

        # Ensure the timing was recorded.
        packet = self.statsd.socket.recv()
        self.assertIsNone(packet)

    def test_timed_start_stop_calls(self):
        # In seconds
        timer = self.statsd.timed('timed_context.test')
        timer.start()
        time.sleep(0.5)
        timer.stop()

        packet = self.recv(2).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test', name)
        self.assert_almost_equal(0.5, float(value), 0.1)

        # In milliseconds
        timer = self.statsd.timed('timed_context.test', use_ms=True)
        timer.start()
        time.sleep(0.5)
        timer.stop()

        packet = self.recv(2, reset_wait=True).split("\n")[0] # ignore telemetry packet
        name_value, type_ = packet.split('|')
        name, value = name_value.split(':')

        self.assertEqual('ms', type_)
        self.assertEqual('timed_context.test', name)
        self.assert_almost_equal(500, float(value), 100)

    def test_batching(self):
        self.statsd.open_buffer()
        self.statsd.gauge('page.views', 123)
        self.statsd.timing('timer', 123)
        self.statsd.close_buffer()
        expected = 'page.views:123|g\ntimer:123|ms\n'
        self.assert_equal_telemetry(
                expected,
                self.recv(2),
                telemetry=telemetry_metrics(metrics=2, bytes_sent=len(expected))
        )

    def test_flush(self):
        dogstatsd = DogStatsd(disable_buffering=False, telemetry_min_flush_interval=0)
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        dogstatsd.increment('page.views')
        self.assertIsNone(fake_socket.recv(no_wait=True))
        dogstatsd.flush()
        self.assert_equal_telemetry('page.views:1|c\n', fake_socket.recv(2))

    def test_flush_interval(self):
        dogstatsd = DogStatsd(disable_buffering=False, flush_interval=1, telemetry_min_flush_interval=0)
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        dogstatsd.increment('page.views')
        self.assertIsNone(fake_socket.recv(no_wait=True))

        time.sleep(0.3)
        self.assertIsNone(fake_socket.recv(no_wait=True))

        time.sleep(1)
        self.assert_equal_telemetry(
            'page.views:1|c\n',
            fake_socket.recv(2, no_wait=True)
        )

    def test_disable_buffering(self):
        dogstatsd = DogStatsd(disable_buffering=True, telemetry_min_flush_interval=0)
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        dogstatsd.increment('page.views')
        self.assert_equal_telemetry(
            'page.views:1|c\n',
            fake_socket.recv(2, no_wait=True)
        )

    def test_flush_disable(self):
        dogstatsd = DogStatsd(
            disable_buffering=False,
            flush_interval=0,
            telemetry_min_flush_interval=0
        )
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        dogstatsd.increment('page.views')
        self.assertIsNone(fake_socket.recv(no_wait=True))

        time.sleep(DEFAULT_FLUSH_INTERVAL)
        self.assertIsNone(fake_socket.recv(no_wait=True))

        time.sleep(0.3)
        self.assertIsNone(fake_socket.recv(no_wait=True))

    @unittest.skip("Buffering has been disabled again so the deprecation is not valid")
    @patch("warnings.warn")
    def test_manual_buffer_ops_deprecation(self, mock_warn):
        self.assertFalse(mock_warn.called)

        self.statsd.open_buffer()
        self.assertTrue(mock_warn.called)
        self.assertEqual(mock_warn.call_count, 1)

        self.statsd.close_buffer()
        self.assertEqual(mock_warn.call_count, 2)

    def test_batching_sequential(self):
        self.statsd.open_buffer()
        self.statsd.gauge('discarded.data', 123)
        self.statsd.close_buffer()

        self.statsd.open_buffer()
        self.statsd.gauge('page.views', 123)
        self.statsd.timing('timer', 123)
        self.statsd.close_buffer()

        expected1 = 'discarded.data:123|g\n'
        expected_metrics1=telemetry_metrics(metrics=1, bytes_sent=len(expected1))
        self.assert_equal_telemetry(
            expected1,
            self.recv(2),
            telemetry=expected_metrics1)


        expected2 = 'page.views:123|g\ntimer:123|ms\n'
        self.assert_equal_telemetry(
            expected2,
            self.recv(2),
            telemetry=telemetry_metrics(
                metrics=2,
                packets_sent=2,
                bytes_sent=len(expected2 + expected_metrics1)
            )
        )

    def test_batching_runtime_changes(self):
        dogstatsd = DogStatsd(
            disable_buffering=True,
            telemetry_min_flush_interval=0
        )
        dogstatsd.socket = FakeSocket()

        # Send some unbuffered metrics and verify we got it immediately
        last_telemetry_size = self.send_and_assert(
            dogstatsd,
            [
                ('gauge', 'rt.gauge', 123),
                ('timing', 'rt.timer', 123),
            ],
        )

        # Disable buffering (noop expected) and validate
        dogstatsd.disable_buffering = True
        last_telemetry_size = self.send_and_assert(
            dogstatsd,
            [
                ('gauge', 'rt.gauge2', 321),
                ('timing', 'rt.timer2', 321),
            ],
            last_telemetry_size = last_telemetry_size,
        )

        # Enable buffering and validate
        dogstatsd.disable_buffering = False
        last_telemetry_size = self.send_and_assert(
            dogstatsd,
            [
                ('gauge', 'buffered.gauge', 12345),
                ('timing', 'buffered.timer', 12345),
            ],
            last_telemetry_size = last_telemetry_size,
            buffered=True,
        )

        # Enable buffering again (another noop change expected)
        dogstatsd.disable_buffering = False
        last_telemetry_size = self.send_and_assert(
            dogstatsd,
            [
                ('gauge', 'buffered.gauge2', 321),
                ('timing', 'buffered.timer2', 321),
            ],
            last_telemetry_size = last_telemetry_size,
            buffered=True,
        )

        # Flip the toggle to unbuffered functionality one more time and verify
        dogstatsd.disable_buffering = True
        last_telemetry_size = self.send_and_assert(
            dogstatsd,
            [
                ('gauge', 'rt.gauge3', 333),
                ('timing', 'rt.timer3', 333),
            ],
            last_telemetry_size = last_telemetry_size,
        )

    def test_threaded_batching(self):
        num_threads = 4
        threads = []

        dogstatsd = DogStatsd(telemetry_min_flush_interval=0)
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        def batch_metrics(index, dsd):
            time.sleep(0.3 * index)

            dsd.open_buffer()

            time.sleep(0.1)
            dsd.gauge('page.%d.views' % index, 123)

            time.sleep(0.1)
            dsd.timing('timer.%d' % index, 123)

            time.sleep(0.5)
            dsd.close_buffer()

        for idx in range(num_threads):
            thread = Thread(
                name="{}_sender_thread_{}".format(self.__class__.__name__, idx),
                target=batch_metrics,
                args=(idx, dogstatsd)
            )
            thread.daemon = True

            threads.append(thread)

        for thread in threads:
            thread.start()

        time.sleep(5)

        for thread in threads:
            if thread.is_alive():
                thread.join(0.1)

        previous_telemetry_packet_size = 0
        thread_idx = 0

        while thread_idx < num_threads:
            first_message = "page.{}.views:123|g\n".format(thread_idx)
            first_message_len = len(first_message)
            second_message = "timer.{}:123|ms\n".format(thread_idx)
            second_message_len = len(second_message)

            received_payload = fake_socket.recv(1)

            # Base assumption is that we got both messages but
            # we may get metrics split depending on when the flush thread triggers
            if received_payload == first_message:
                message = first_message
                packet_size = first_message_len
                num_metrics = 1
            elif received_payload == second_message:
                message = second_message
                packet_size = second_message_len
                num_metrics = 1
                thread_idx += 1
            else:
                message = first_message + second_message
                packet_size = len(message)
                num_metrics = 2
                thread_idx += 1

            self.assertEqual(received_payload, message)

            packet_sent = 2
            if previous_telemetry_packet_size == 0:
                packet_sent = 1

            bytes_sent = previous_telemetry_packet_size + packet_size
            telemetry = telemetry_metrics(
                    metrics=num_metrics,
                    bytes_sent=bytes_sent,
                    packets_sent=packet_sent,
            )
            self.assertEqual(telemetry, fake_socket.recv(1))

            previous_telemetry_packet_size = len(telemetry)

    def test_telemetry(self):
        self.statsd.metrics_count = 1
        self.statsd.events_count = 2
        self.statsd.service_checks_count = 3
        self.statsd.bytes_sent = 4
        self.statsd.bytes_dropped = 5
        self.statsd.packets_sent = 6
        self.statsd.packets_dropped = 7

        self.statsd.open_buffer()
        self.statsd.gauge('page.views', 123)
        self.statsd.close_buffer()

        payload = 'page.views:123|g\n'
        telemetry = telemetry_metrics(metrics=2, events=2, service_checks=3, bytes_sent=4 + len(payload),
                                      bytes_dropped=5, packets_sent=7, packets_dropped=7)

        self.assert_equal_telemetry(payload, self.recv(2), telemetry=telemetry)

        self.assertEqual(0, self.statsd.metrics_count)
        self.assertEqual(0, self.statsd.events_count)
        self.assertEqual(0, self.statsd.service_checks_count)
        self.assertEqual(len(telemetry), self.statsd.bytes_sent)
        self.assertEqual(0, self.statsd.bytes_dropped)
        self.assertEqual(1, self.statsd.packets_sent)
        self.assertEqual(0, self.statsd.packets_dropped)

    def test_telemetry_flush_interval(self):
        dogstatsd = DogStatsd(disable_buffering=False)
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        # Set the last flush time in the future to be sure we won't flush
        dogstatsd._last_flush_time = time.time() + dogstatsd._telemetry_flush_interval
        dogstatsd.gauge('gauge', 123.4)

        metric = 'gauge:123.4|g\n'
        self.assertEqual(metric, fake_socket.recv())

        time1 = time.time()
        # Setting the last flush time in the past to trigger a telemetry flush
        dogstatsd._last_flush_time = time1 - dogstatsd._telemetry_flush_interval -1
        dogstatsd.gauge('gauge', 123.4)
        self.assert_equal_telemetry(
            metric,
            fake_socket.recv(2, reset_wait=True),
            telemetry=telemetry_metrics(
                metrics=2,
                bytes_sent=2*len(metric),
                packets_sent=2,
            ),
        )

        # assert that _last_flush_time has been updated
        self.assertTrue(time1 < dogstatsd._last_flush_time)

    def test_telemetry_flush_interval_alternate_destination(self):
        dogstatsd = DogStatsd(telemetry_host='foo')
        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket
        fake_telemetry_socket = FakeSocket()
        dogstatsd.telemetry_socket = fake_telemetry_socket

        self.assertIsNotNone(dogstatsd.telemetry_host)
        self.assertIsNotNone(dogstatsd.telemetry_port)
        self.assertTrue(dogstatsd._dedicated_telemetry_destination())

        # set the last flush time in the future to be sure we won't flush
        dogstatsd._last_flush_time = time.time() + dogstatsd._telemetry_flush_interval
        dogstatsd.gauge('gauge', 123.4)

        self.assertEqual('gauge:123.4|g\n', fake_socket.recv())

        time1 = time.time()
        # setting the last flush time in the past to trigger a telemetry flush
        dogstatsd._last_flush_time = time1 - dogstatsd._telemetry_flush_interval - 1
        dogstatsd.gauge('gauge', 123.4)

        self.assertEqual('gauge:123.4|g\n', fake_socket.recv(reset_wait=True))
        self.assert_equal_telemetry(
            '',
            fake_telemetry_socket.recv(),
            telemetry=telemetry_metrics(
                metrics=2,
                bytes_sent=14*2,
                packets_sent=2,
            ),
        )

        # assert that _last_flush_time has been updated
        self.assertTrue(time1 < dogstatsd._last_flush_time)

    def test_telemetry_flush_interval_batch(self):
        dogstatsd = DogStatsd(disable_buffering=False)

        fake_socket = FakeSocket()
        dogstatsd.socket = fake_socket

        dogstatsd.open_buffer()
        dogstatsd.gauge('gauge1', 1)
        dogstatsd.gauge('gauge2', 2)

        time1 = time.time()
        # setting the last flush time in the past to trigger a telemetry flush
        dogstatsd._last_flush_time = time1 - statsd._telemetry_flush_interval -1

        dogstatsd.close_buffer()

        metric = 'gauge1:1|g\ngauge2:2|g\n'
        self.assert_equal_telemetry(metric, fake_socket.recv(2), telemetry=telemetry_metrics(metrics=2, bytes_sent=len(metric)))
        # assert that _last_flush_time has been updated
        self.assertTrue(time1 < dogstatsd._last_flush_time)


    def test_dedicated_udp_telemetry_dest(self):
        listener_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        listener_sock.bind(('localhost', 0))

        def wait_for_data():
            global udp_thread_telemetry_data
            udp_thread_telemetry_data = listener_sock.recvfrom(UDP_OPTIMAL_PAYLOAD_LENGTH)[0].decode('utf-8')

        with closing(listener_sock):
            port = listener_sock.getsockname()[1]

            dogstatsd = DogStatsd(
                host="localhost",
                port=12345,
                telemetry_min_flush_interval=0,
                telemetry_host="localhost",
                telemetry_port=port,
            )

            server = threading.Thread(target=wait_for_data)
            server.start()

            dogstatsd.increment('abc')

            server.join(3)

            expected_telemetry = telemetry_metrics(metrics=1, packets_sent=1, bytes_sent=8)
            self.assertEqual(udp_thread_telemetry_data, expected_telemetry)

    def test_dedicated_uds_telemetry_dest(self):
        tempdir = tempfile.mkdtemp()
        socket_path = os.path.join(tempdir, 'socket.sock')

        listener_sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        listener_sock.bind(socket_path)

        def wait_for_data():
            global uds_thread_telemetry_data
            uds_thread_telemetry_data = listener_sock.recvfrom(UDS_OPTIMAL_PAYLOAD_LENGTH)[0].decode('utf-8')

        with closing(listener_sock):
            dogstatsd = DogStatsd(
                host="localhost",
                port=12345,
                telemetry_min_flush_interval=0,
                telemetry_socket_path=socket_path,
            )

            server = threading.Thread(target=wait_for_data)
            server.start()

            dogstatsd.increment('def')

            server.join(3)

            expected_telemetry = telemetry_metrics(metrics=1, packets_sent=1, bytes_sent=8)
            self.assertEqual(uds_thread_telemetry_data, expected_telemetry)

        shutil.rmtree(tempdir)

    def test_context_manager(self):
        fake_socket = FakeSocket()
        with DogStatsd(telemetry_min_flush_interval=0) as dogstatsd:
            dogstatsd.socket = fake_socket
            dogstatsd.gauge('page.views', 123)
            dogstatsd.timing('timer', 123)
            dogstatsd.increment('my_counter', 3)

        metric1 = "page.views:123|g"
        metric2 = "timer:123|ms"
        metric3 = "my_counter:3|c"

        metrics = '\n'.join([metric1, metric2, metric3]) + "\n"
        self.assertEqual(metrics, fake_socket.recv(no_wait=True))

        metrics_packet = telemetry_metrics(
            metrics=3,
            bytes_sent=len(metrics),
            packets_sent=1,
        )
        self.assertEqual(metrics_packet, fake_socket.recv(no_wait=True))

    def test_batched_buffer_autoflush(self):
        fake_socket = FakeSocket()
        bytes_sent = 0
        with DogStatsd(telemetry_min_flush_interval=0, disable_buffering=False) as dogstatsd:
            dogstatsd.socket = fake_socket

            self.assertEqual(dogstatsd._max_payload_size, UDP_OPTIMAL_PAYLOAD_LENGTH)

            single_metric = 'mycounter:1|c\n'
            metrics_per_packet = dogstatsd._max_payload_size // len(single_metric)
            for _ in range(metrics_per_packet + 1):
                dogstatsd.increment('mycounter')
            payload = ''.join([single_metric for _ in range(metrics_per_packet)])

            telemetry = telemetry_metrics(
                metrics=metrics_per_packet+1,
                bytes_sent=len(payload),
            )
            bytes_sent += len(payload) + len(telemetry)
            self.assertEqual(payload, fake_socket.recv())
            self.assertEqual(telemetry, fake_socket.recv())

        self.assertEqual(single_metric, fake_socket.recv())

        telemetry = telemetry_metrics(metrics=0, packets_sent=2, bytes_sent=len(single_metric) + len(telemetry))
        self.assertEqual(telemetry, fake_socket.recv())

    def test_module_level_instance(self):
        self.assertTrue(isinstance(statsd, DogStatsd))

    def test_instantiating_does_not_connect(self):
        dogpound = DogStatsd()
        self.assertIsNone(dogpound.socket)

    def test_accessing_socket_opens_socket(self):
        dogpound = DogStatsd()
        try:
            self.assertIsNotNone(dogpound.get_socket())
        finally:
            dogpound.socket.close()

    def test_accessing_socket_multiple_times_returns_same_socket(self):
        dogpound = DogStatsd()
        fresh_socket = FakeSocket()
        dogpound.socket = fresh_socket
        self.assertEqual(fresh_socket, dogpound.get_socket())
        self.assertNotEqual(FakeSocket(), dogpound.get_socket())

    def test_tags_from_environment(self):
        with preserve_environment_variable('DATADOG_TAGS'):
            os.environ['DATADOG_TAGS'] = 'country:china,age:45,blue'
            dogstatsd = DogStatsd(telemetry_min_flush_interval=0)
        dogstatsd.socket = FakeSocket()
        dogstatsd.gauge('gt', 123.4)
        metric = 'gt:123.4|g|#country:china,age:45,blue\n'
        self.assertEqual(metric, dogstatsd.socket.recv())
        self.assertEqual(telemetry_metrics(tags="country:china,age:45,blue", bytes_sent=len(metric)), dogstatsd.socket.recv())

    def test_tags_from_environment_and_constant(self):
        with preserve_environment_variable('DATADOG_TAGS'):
            os.environ['DATADOG_TAGS'] = 'country:china,age:45,blue'
            dogstatsd = DogStatsd(constant_tags=['country:canada', 'red'], telemetry_min_flush_interval=0)
        dogstatsd.socket = FakeSocket()
        dogstatsd.gauge('gt', 123.4)
        tags = "country:canada,red,country:china,age:45,blue"
        metric = 'gt:123.4|g|#' + tags + '\n'
        self.assertEqual(metric, dogstatsd.socket.recv())
        self.assertEqual(telemetry_metrics(tags=tags, bytes_sent=len(metric)), dogstatsd.socket.recv())

    def test_entity_tag_from_environment(self):
        with preserve_environment_variable('DD_ENTITY_ID'):
            os.environ['DD_ENTITY_ID'] = '04652bb7-19b7-11e9-9cc6-42010a9c016d'
            dogstatsd = DogStatsd(telemetry_min_flush_interval=0)
        dogstatsd.socket = FakeSocket()
        dogstatsd.gauge('gt', 123.4)
        metric = 'gt:123.4|g|#dd.internal.entity_id:04652bb7-19b7-11e9-9cc6-42010a9c016d\n'
        self.assertEqual(metric, dogstatsd.socket.recv())
        self.assertEqual(
            telemetry_metrics(tags="dd.internal.entity_id:04652bb7-19b7-11e9-9cc6-42010a9c016d", bytes_sent=len(metric)),
            dogstatsd.socket.recv())

    def test_entity_tag_from_environment_and_constant(self):
        with preserve_environment_variable('DD_ENTITY_ID'):
            os.environ['DD_ENTITY_ID'] = '04652bb7-19b7-11e9-9cc6-42010a9c016d'
            dogstatsd = DogStatsd(constant_tags=['country:canada', 'red'], telemetry_min_flush_interval=0)
        dogstatsd.socket = FakeSocket()
        dogstatsd.gauge('gt', 123.4)
        metric = 'gt:123.4|g|#country:canada,red,dd.internal.entity_id:04652bb7-19b7-11e9-9cc6-42010a9c016d\n'
        self.assertEqual(metric, dogstatsd.socket.recv())
        self.assertEqual(
            telemetry_metrics(tags="country:canada,red,dd.internal.entity_id:04652bb7-19b7-11e9-9cc6-42010a9c016d",
                              bytes_sent=len(metric)),
            dogstatsd.socket.recv()
        )

    def test_entity_tag_and_tags_from_environment_and_constant(self):
        with preserve_environment_variable('DATADOG_TAGS'):
            os.environ['DATADOG_TAGS'] = 'country:china,age:45,blue'
            with preserve_environment_variable('DD_ENTITY_ID'):
                os.environ['DD_ENTITY_ID'] = '04652bb7-19b7-11e9-9cc6-42010a9c016d'
                dogstatsd = DogStatsd(constant_tags=['country:canada', 'red'], telemetry_min_flush_interval=0)
        dogstatsd.socket = FakeSocket()
        dogstatsd.gauge('gt', 123.4)
        tags = "country:canada,red,country:china,age:45,blue,dd.internal.entity_id:04652bb7-19b7-11e9-9cc6-42010a9c016d"
        metric = 'gt:123.4|g|#' + tags + '\n'
        self.assertEqual(metric, dogstatsd.socket.recv())
        self.assertEqual(telemetry_metrics(tags=tags, bytes_sent=len(metric)), dogstatsd.socket.recv())

    def test_dogstatsd_initialization_with_dd_env_service_version(self):
        """
        Dogstatsd should automatically use DD_ENV, DD_SERVICE, and DD_VERSION (if present)
        to set {env, service, version} as global tags for all metrics emitted.
        """
        cases = [
            # Test various permutations of setting DD_* env vars, as well as other global tag configuration.
            # An empty string signifies that the env var either isn't set or that it is explicitly set to empty string.
            ('', '', '', '', [], []),
            ('prod', '', '', '', [], ['env:prod']),
            ('prod', 'dog', '', '', [], ['env:prod', 'service:dog']),
            ('prod', 'dog', 'abc123', '', [], ['env:prod', 'service:dog', 'version:abc123']),
            ('prod', 'dog', 'abc123', 'env:prod,type:app', [], ['env:prod', 'env:prod', 'service:dog', 'type:app', 'version:abc123']),
            ('prod', 'dog', 'abc123', 'env:prod2,type:app', [], ['env:prod', 'env:prod2', 'service:dog', 'type:app', 'version:abc123']),
            ('prod', 'dog', 'abc123', '', ['env:prod', 'type:app'], ['env:prod', 'env:prod', 'service:dog', 'type:app', 'version:abc123']),
            ('prod', 'dog', 'abc123', '', ['env:prod2', 'type:app'], ['env:prod', 'env:prod2', 'service:dog', 'type:app', 'version:abc123']),
            ('prod', 'dog', 'abc123', 'env:prod3,custom_tag:cat', ['env:prod2', 'type:app'], ['custom_tag:cat', 'env:prod', 'env:prod2', 'env:prod3', 'service:dog', 'type:app', 'version:abc123']),
        ]
        for case in cases:
            dd_env, dd_service, dd_version, datadog_tags, constant_tags, global_tags = case
            with EnvVars(
                env_vars={
                    'DATADOG_TAGS': datadog_tags,
                    'DD_ENV': dd_env,
                    'DD_SERVICE': dd_service,
                    'DD_VERSION': dd_version,
                }
            ):
                dogstatsd = DogStatsd(constant_tags=constant_tags, telemetry_min_flush_interval=0)
                dogstatsd.socket = FakeSocket()

            # Guarantee consistent ordering, regardless of insertion order.
            dogstatsd.constant_tags.sort()
            self.assertEqual(global_tags, dogstatsd.constant_tags)

            # Make call with no tags passed; only the globally configured tags will be used.
            global_tags_str = ','.join([t for t in global_tags])
            dogstatsd.gauge('gt', 123.4)
            dogstatsd.flush()

            # Protect against the no tags case.
            metric = 'gt:123.4|g|#{}\n'.format(global_tags_str) if global_tags_str else 'gt:123.4|g\n'
            self.assertEqual(metric, dogstatsd.socket.recv())
            self.assertEqual(
                telemetry_metrics(
                    tags=global_tags_str,
                    bytes_sent=len(metric)
                ),
                dogstatsd.socket.recv(),
            )
            dogstatsd._reset_telemetry()

            # Make another call with local tags passed.
            passed_tags = ['env:prod', 'version:def456', 'custom_tag:toad']
            all_tags_str = ','.join([t for t in passed_tags + global_tags])
            dogstatsd.gauge('gt', 123.4, tags=passed_tags)
            dogstatsd.flush()

            metric = 'gt:123.4|g|#{}\n'.format(all_tags_str)
            self.assertEqual(metric, dogstatsd.socket.recv())
            self.assertEqual(
                telemetry_metrics(
                    tags=global_tags_str,
                    bytes_sent=len(metric),
                ),
                dogstatsd.socket.recv(),
            )

    def test_default_max_udp_packet_size(self):
        dogstatsd = DogStatsd(disable_buffering=False, flush_interval=10000, disable_telemetry=True)
        dogstatsd.socket = FakeSocket()

        for _ in range(10000):
            dogstatsd.increment('val')

        payload = dogstatsd.socket.recv()
        self.assertIsNotNone(payload)
        while payload is not None:
            payload_size = len(payload)
            self.assertLessEqual(payload_size, UDP_OPTIMAL_PAYLOAD_LENGTH)
            self.assertGreater(payload_size, UDP_OPTIMAL_PAYLOAD_LENGTH - 100)

            payload = dogstatsd.socket.recv()

    def test_default_max_uds_packet_size(self):
        dogstatsd = DogStatsd(
            disable_buffering=False,
            socket_path="fake",
            flush_interval=10000,
            disable_telemetry=True,
        )
        dogstatsd.socket = FakeSocket()

        for _ in range(10000):
            dogstatsd.increment('val')

        payload = dogstatsd.socket.recv()
        self.assertIsNotNone(payload)
        while payload is not None:
            payload_size = len(payload)
            self.assertLessEqual(payload_size, UDS_OPTIMAL_PAYLOAD_LENGTH)
            self.assertGreater(payload_size, UDS_OPTIMAL_PAYLOAD_LENGTH - 100)

            payload = dogstatsd.socket.recv()

    def test_custom_max_packet_size(self):
        dogstatsd = DogStatsd(
            disable_buffering=False,
            max_buffer_len=4000,
            flush_interval=10000,
            disable_telemetry=True,
        )
        dogstatsd.socket = FakeSocket()

        for _ in range(10000):
            dogstatsd.increment('val')

        payload = dogstatsd.socket.recv()
        self.assertIsNotNone(payload)
        while payload is not None:
            payload_size = len(payload)
            self.assertLessEqual(payload_size, 4000)
            self.assertGreater(payload_size, 3900)

            payload = dogstatsd.socket.recv()

    def test_gauge_does_not_send_none(self):
        self.statsd.gauge('metric', None)
        self.assertIsNone(self.recv())

    def test_increment_does_not_send_none(self):
        self.statsd.increment('metric', None)
        self.assertIsNone(self.recv())

    def test_decrement_does_not_send_none(self):
        self.statsd.decrement('metric', None)
        self.assertIsNone(self.recv())

    def test_timing_does_not_send_none(self):
        self.statsd.timing('metric', None)
        self.assertIsNone(self.recv())

    def test_histogram_does_not_send_none(self):
        self.statsd.histogram('metric', None)
        self.assertIsNone(self.recv())



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from argparse import ArgumentTypeError
from freezegun import freeze_time
import datetime
import unittest

from datadog.util.cli import (
    comma_list,
    comma_set,
    comma_list_or_empty,
    list_of_ints,
    list_of_ints_and_strs,
    set_of_ints,
    DateParsingError,
    _midnight,
    parse_date_as_epoch_timestamp,
    parse_date,
)
from datadog.util.compat import is_pypy
from datadog.util.format import force_to_epoch_seconds


class TestCLI(unittest.TestCase):
    def test_comma_list(self):
        invalid_cases = [None, ""]
        for invalid_case in invalid_cases:
            with self.assertRaises(ArgumentTypeError):
                comma_list(invalid_case)

        valid_cases = (
            (["foo"], "foo", None),
            (["foo", "bar"], "foo,bar", None),
            ([1], "1", int),
            ([1, 2], "1,2", int),
        )
        for expected, list_str, item_func in valid_cases:
            actual = comma_list(list_str, item_func)
            self.assertListEqual(expected, actual)

    def test_comma_set(self):
        invalid_cases = [None, ""]
        for invalid_case in invalid_cases:
            with self.assertRaises(ArgumentTypeError):
                comma_set(invalid_case)

        valid_cases = (
            ({"foo"}, "foo", None),
            ({"foo", "bar"}, "foo,bar", None),
            ({1}, "1", int),
            ({1}, "1,1,1", int),
            ({1, 2}, "1,2,1", int),
        )
        for expected, list_str, item_func in valid_cases:
            actual = comma_set(list_str, item_func)
            self.assertSetEqual(expected, actual)

    def test_comma_list_or_empty(self):
        valid_cases = (
            ([], None, None),
            ([], "", None),
            (["foo"], "foo", None),
            (["foo", "bar"], "foo,bar", None),
        )
        for expected, list_str, item_func in valid_cases:
            actual = comma_list_or_empty(list_str)
            self.assertListEqual(expected, actual)

    def test_list_of_ints(self):
        invalid_cases = [None, "", "foo", '["foo"]']
        for invalid_case in invalid_cases:
            with self.assertRaises(ArgumentTypeError):
                list_of_ints(invalid_case)

        valid_cases = (([1], "1"), ([1, 2], "1,2"), ([1], "[1]"), ([1, 2], "[1,2]"))
        for expected, list_str in valid_cases:
            actual = list_of_ints(list_str)
            self.assertListEqual(expected, actual)

    def test_list_of_ints_and_strs(self):
        invalid_cases = [None, ""]
        for invalid_case in invalid_cases:
            with self.assertRaises(ArgumentTypeError):
                list_of_ints_and_strs(invalid_case)

        valid_cases = (
            (["foo"], "foo"),
            (["foo", "bar"], "foo,bar"),
            ([1], "1"),
            ([1, 2], "1,2"),
            (["foo", 2], "foo,2"),
        )
        for expected, list_str in valid_cases:
            actual = list_of_ints_and_strs(list_str)
            self.assertListEqual(expected, actual)

    def test_set_of_ints(self):
        invalid_cases = [None, "", "foo", '["foo"]']
        for invalid_case in invalid_cases:
            with self.assertRaises(ArgumentTypeError):
                set_of_ints(invalid_case)

        valid_cases = (
            ({1}, "1"),
            ({1, 2}, "1,2"),
            ({1}, "[1]"),
            ({1}, "[1,1,1]"),
            ({1, 2}, "[1,2,1]"),
        )
        for expected, list_str in valid_cases:
            actual = set_of_ints(list_str)
            self.assertSetEqual(expected, actual)

    @freeze_time("2019-10-23 04:44:32", tz_offset=0)
    def test_midnight(self):
        d = _midnight()
        self.assertEqual(2019, d.year)
        self.assertEqual(10, d.month)
        self.assertEqual(23, d.day)
        self.assertEqual(0, d.hour)
        self.assertEqual(0, d.minute)
        self.assertEqual(0, d.second)
        self.assertEqual(0, d.microsecond)

    @freeze_time("2019-10-23 04:44:32", tz_offset=0)
    def test_parse_date(self):
        test_date = datetime.datetime(2019, 10, 23, 4, 44, 32, 0)
        cases = [
            (test_date, test_date),  # already an instance, return
            ("today", datetime.datetime(2019, 10, 23, 0, 0, 0)),
            ("yesterday", datetime.datetime(2019, 10, 22, 0, 0, 0)),
            ("tomorrow", datetime.datetime(2019, 10, 24, 0, 0, 0)),
            ("2 days ago", datetime.datetime(2019, 10, 21, 4, 44, 32)),
            ("2d ago", datetime.datetime(2019, 10, 21, 4, 44, 32)),
            ("2 days ahead", datetime.datetime(2019, 10, 25, 4, 44, 32)),
            ("2d ahead", datetime.datetime(2019, 10, 25, 4, 44, 32)),
            ("now", datetime.datetime(2019, 10, 23, 4, 44, 32)),
            ("2019-10-23 04:44:32.000000", test_date),
            ("2019-10-23T04:44:32.000000", test_date),
            ("2019-10-23 04:44:32", test_date),
            ("2019-10-23T04:44:32", test_date),
            ("2019-10-23 04:44", datetime.datetime(2019, 10, 23, 4, 44, 0, 0)),
            ("2019-10-23-04", datetime.datetime(2019, 10, 23, 4, 0, 0, 0)),
            ("2019-10-23", datetime.datetime(2019, 10, 23, 0, 0, 0, 0)),
            ("2019-10", datetime.datetime(2019, 10, 1, 0, 0, 0, 0)),
            ("2019", datetime.datetime(2019, 1, 1, 0, 0, 0, 0)),
            ("2019-10", datetime.datetime(2019, 10, 1, 0, 0, 0, 0)),
            ("1571805872", test_date),  # seconds
        ]
        if not is_pypy():
            cases.append(
                ("1571805872000", test_date)
            )  # millis, pypy does not work (known)

        for i, (date_str, expected) in enumerate(cases):
            actual = parse_date(date_str)
            self.assertEqual(
                expected,
                actual,
                "case {}: failed, date_str={} expected={} actual={}".format(
                    i, date_str, expected, actual
                ),
            )

        # test invalid case
        with self.assertRaises(DateParsingError):
            parse_date("foo")

    @freeze_time("2019-10-23 04:44:32", tz_offset=0)
    def test_parse_date_as_epoch_timestamp(self):
        # this applies the same rules but always returns epoch seconds
        test_date = datetime.datetime(2019, 10, 23, 4, 44, 32, 0)
        cases = [
            (test_date, test_date),  # already an instance, return
            ("today", datetime.datetime(2019, 10, 23, 0, 0, 0)),
            ("yesterday", datetime.datetime(2019, 10, 22, 0, 0, 0)),
            ("tomorrow", datetime.datetime(2019, 10, 24, 0, 0, 0)),
            ("2 days ago", datetime.datetime(2019, 10, 21, 4, 44, 32)),
            ("2d ago", datetime.datetime(2019, 10, 21, 4, 44, 32)),
            ("2 days ahead", datetime.datetime(2019, 10, 25, 4, 44, 32)),
            ("2d ahead", datetime.datetime(2019, 10, 25, 4, 44, 32)),
            ("now", datetime.datetime(2019, 10, 23, 4, 44, 32)),
            ("2019-10-23 04:44:32.000000", test_date),
            ("2019-10-23T04:44:32.000000", test_date),
            ("2019-10-23 04:44:32", test_date),
            ("2019-10-23T04:44:32", test_date),
            ("2019-10-23 04:44", datetime.datetime(2019, 10, 23, 4, 44, 0, 0)),
            ("2019-10-23-04", datetime.datetime(2019, 10, 23, 4, 0, 0, 0)),
            ("2019-10-23", datetime.datetime(2019, 10, 23, 0, 0, 0, 0)),
            ("2019-10", datetime.datetime(2019, 10, 1, 0, 0, 0, 0)),
            ("2019", datetime.datetime(2019, 1, 1, 0, 0, 0, 0)),
            ("2019-10", datetime.datetime(2019, 10, 1, 0, 0, 0, 0)),
            ("1571805872", test_date),  # seconds
        ]
        if not is_pypy():
            cases.append(
                ("1571805872000", test_date)
            )  # millis, pypy does not work (known)

        for i, (date_str, expected) in enumerate(cases):
            actual_timestamp = parse_date_as_epoch_timestamp(date_str)
            expected_timestamp = force_to_epoch_seconds(expected)
            self.assertEqual(
                expected_timestamp,
                actual_timestamp,
                "case {}: failed, date_str={} expected={} actual={}".format(
                    i, date_str, expected_timestamp, actual_timestamp
                ),
            )

        # test invalid case
        with self.assertRaises(DateParsingError):
            parse_date_as_epoch_timestamp("foo")



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
"""
Tests for the ThreadStats class, using HTTP mode
"""

# stdlib
import logging
import os
import random
import time
import unittest

# 3p
from mock import patch

# datadog
from datadog import ThreadStats, lambda_metric, datadog_lambda_wrapper
from datadog.threadstats.aws_lambda import _get_lambda_stats
from tests.util.contextmanagers import preserve_environment_variable, EnvVars

# Silence the logger.
logger = logging.getLogger('dd.datadogpy')
logger.setLevel(logging.ERROR)


class MemoryReporter(object):
    """
    A reporting class that reports to memory for testing.
    """

    def __init__(self):
        self.distributions = []
        self.metrics = []
        self.events = []
        self.dist_flush_counter = 0

    def flush_distributions(self, distributions):
        self.distributions += distributions
        self.dist_flush_counter = self.dist_flush_counter + 1

    def flush_metrics(self, metrics):
        self.metrics += metrics

    def flush_events(self, events):
        self.events += events


class TestUnitThreadStats(unittest.TestCase):
    """
    Unit tests for ThreadStats.
    """
    def setUp(self):
        """
        Set a mocked reporter.
        """
        self.reporter = MemoryReporter()

    def sort_metrics(self, metrics):
        """
        Sort metrics by timestamp of first point and then name.
        """
        def sort(metric):
            tags = metric['tags'] or []
            host = metric['host'] or ''
            return (metric['points'][0][0], metric['metric'], tags, host,
                    metric['points'][0][1])
        return sorted(metrics, key=sort)

    def assertMetric(self, name=None, value=None, tags=None, count=None):
        """
        Helper, to make assertions on metrics.
        """
        matching_metrics = []

        for metric in self.reporter.metrics:
            if name and name != metric['metric']:
                continue
            if value and value != metric['points'][0][1]:
                continue
            if tags and tags != metric['tags']:
                continue
            matching_metrics.append(metric)

        if count:
            self.assertEqual(
                len(matching_metrics), count,
                u"Candidate size assertion failure: expected {expected}, found {count}. "
                u"Metric name={name}, value={value}, tags={tags}.".format(
                    expected=count, count=len(matching_metrics),
                    name=name, value=value, tags=tags
                )
            )
        else:
            self.assertTrue(
                len(matching_metrics) > 0,
                u"Candidate size assertion failure: no matching metric found. "
                u"Metric name={name}, value={value}, tags={tags}.".format(
                    name=name, value=value, tags=tags
                )
            )

    def test_init(self):
        # Test compress_payload setting
        t = ThreadStats(compress_payload=True)
        t.start()
        assert t.reporter.compress_payload is True
        t.stop()
        # Default value
        t = ThreadStats()
        t.start()
        assert t.reporter.compress_payload is False
        t.stop()

    def test_timed_decorator(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=1, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        @dog.timed('timed.test')
        def func(a, b, c=1, d=1):
            """docstring"""
            return (a, b, c, d)

        assert func.__name__ == 'func'
        assert func.__doc__ == 'docstring'

        result = func(1, 2, d=3)
        # Assert it handles args and kwargs correctly.
        assert result == (1, 2, 1, 3)
        time.sleep(1)  # Argh. I hate this.
        dog.flush()
        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 8
        (_, _, _, _, avg, count, max_, min_) = metrics
        assert avg['metric'] == 'timed.test.avg'
        assert count['metric'] == 'timed.test.count'
        assert max_['metric'] == 'timed.test.max'
        assert min_['metric'] == 'timed.test.min'

    def test_event(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Add two events
        event1_title = "Event 1 title"
        event2_title = "Event 1 title"
        event1_text = "Event 1 text"
        event2_text = "Event 2 text"
        dog.event(event1_title, event1_text)
        dog.event(event2_title, event2_text)

        # Flush and test
        dog.flush()
        event1, event2 = reporter.events
        assert event1['title'] == event1_title
        assert event1['text'] == event1_text
        assert event2['title'] == event2_title
        assert event2['text'] == event2_text

        # Test more parameters
        reporter.events = []
        event1_priority = "low"
        event1_date_happened = 1375296969
        event1_tag = "Event 2 tag"
        dog.event(event1_title, event1_text, priority=event1_priority,
                  date_happened=event1_date_happened, tags=[event1_tag])

        # Flush and test
        dog.flush()
        event, = reporter.events
        assert event['title'] == event1_title
        assert event['text'] == event1_text
        assert event['priority'] == event1_priority
        assert event['date_happened'] == event1_date_happened
        assert event['tags'] == [event1_tag]

    def test_event_constant_tags(self):
        constant_tag = 'type:constant'
        dog = ThreadStats(constant_tags=[constant_tag])
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Add two events
        event1_title = "Event 1 title"
        event2_title = "Event 1 title"
        event1_text = "Event 1 text"
        event2_text = "Event 2 text"
        dog.event(event1_title, event1_text)
        dog.event(event2_title, event2_text)

        # Flush and test
        dog.flush()
        event1, event2 = reporter.events
        assert event1['title'] == event1_title
        assert event1['text'] == event1_text
        assert event1['tags'] == [constant_tag]
        assert event2['title'] == event2_title
        assert event2['text'] == event2_text
        assert event2['text'] == event2_text
        assert event2['tags'] == [constant_tag]

        # Test more parameters
        reporter.events = []
        event1_priority = "low"
        event1_date_happened = 1375296969
        event1_tag = "Event 2 tag"
        dog.event(event1_title, event1_text, priority=event1_priority,
                  date_happened=event1_date_happened, tags=[event1_tag])

        # Flush and test
        dog.flush()
        event, = reporter.events
        assert event['title'] == event1_title
        assert event['text'] == event1_text
        assert event['priority'] == event1_priority
        assert event['date_happened'] == event1_date_happened
        assert event['tags'] == [event1_tag, constant_tag]

    def test_histogram(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Add some histogram metrics.
        dog.histogram('histogram.1', 20, 100.0)
        dog.histogram('histogram.1', 30, 105.0)
        dog.histogram('histogram.1', 40, 106.0)
        dog.histogram('histogram.1', 50, 106.0)

        dog.histogram('histogram.1', 30, 110.0)
        dog.histogram('histogram.1', 50, 115.0)
        dog.histogram('histogram.1', 40, 116.0)

        dog.histogram('histogram.2', 40, 100.0)

        dog.histogram('histogram.3', 50, 134.0)

        # Flush and ensure they roll up properly.
        dog.flush(120.0)
        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 24

        # Test histograms elsewhere.
        (h1751, h1851, h1951, h1991, h1avg1, h1cnt1, h1max1, h1min1,
         _, _, _, _, h2avg1, h2cnt1, h2max1, h2min1,
         h1752, _, _, h1992, h1avg2, h1cnt2, h1max2, h1min2) = metrics

        assert h1avg1['metric'] == 'histogram.1.avg'
        assert h1avg1['points'][0][0] == 100.0
        assert h1avg1['points'][0][1] == 35
        assert h1cnt1['metric'] == 'histogram.1.count'
        assert h1cnt1['points'][0][0] == 100.0
        assert h1cnt1['points'][0][1] == 0.4
        assert h1min1['metric'] == 'histogram.1.min'
        assert h1min1['points'][0][1] == 20
        assert h1max1['metric'] == 'histogram.1.max'
        assert h1max1['points'][0][1] == 50
        assert h1751['metric'] == 'histogram.1.75percentile'
        assert h1751['points'][0][1] == 40
        assert h1991['metric'] == 'histogram.1.99percentile'
        assert h1991['points'][0][1] == 50

        assert h1avg2['metric'] == 'histogram.1.avg'
        assert h1avg2['points'][0][0] == 110.0
        assert h1avg2['points'][0][1] == 40
        assert h1cnt2['metric'] == 'histogram.1.count'
        assert h1cnt2['points'][0][0] == 110.0
        assert h1cnt2['points'][0][1] == 0.3
        assert h1752['metric'] == 'histogram.1.75percentile'
        assert h1752['points'][0][0] == 110.0
        assert h1752['points'][0][1] == 40.0
        assert h1992['metric'] == 'histogram.1.99percentile'
        assert h1992['points'][0][0] == 110.0
        assert h1992['points'][0][1] == 50.0

        assert h2avg1['metric'] == 'histogram.2.avg'
        assert h2avg1['points'][0][0] == 100.0
        assert h2avg1['points'][0][1] == 40
        assert h2cnt1['metric'] == 'histogram.2.count'
        assert h2cnt1['points'][0][0] == 100.0
        assert h2cnt1['points'][0][1] == 0.1

        # Flush again ensure they're gone.
        dog.reporter.metrics = []
        dog.flush(140.0)
        assert len(dog.reporter.metrics) == 8
        dog.reporter.metrics = []
        dog.flush(200.0)
        assert len(dog.reporter.metrics) == 0

    def test_histogram_percentiles(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()
        # Sample all numbers between 1-100 many times. This
        # means our percentiles should be relatively close to themselves.
        percentiles = list(range(100))
        random.shuffle(percentiles)  # in place
        for i in percentiles:
            for j in range(20):
                dog.histogram('percentiles', i, 1000.0)
        dog.flush(2000.0)
        metrics = reporter.metrics

        def assert_almost_equal(i, j, e=1):
            # Floating point math?
            assert abs(i - j) <= e, "%s %s %s" % (i, j, e)
        assert len(metrics) == 8
        p75, p85, p95, p99, _, _, _, _ = self.sort_metrics(metrics)
        assert p75['metric'] == 'percentiles.75percentile'
        assert p75['points'][0][0] == 1000.0
        assert_almost_equal(p75['points'][0][1], 75, 8)
        assert_almost_equal(p85['points'][0][1], 85, 8)
        assert_almost_equal(p95['points'][0][1], 95, 8)
        assert_almost_equal(p99['points'][0][1], 99, 8)

    def test_gauge(self):
        # Create some fake metrics.
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        dog.gauge('test.gauge.1', 20, 100.0)
        dog.gauge('test.gauge.1', 22, 105.0)
        dog.gauge('test.gauge.2', 30, 115.0)
        dog.gauge('test.gauge.3', 30, 125.0)
        dog.flush(120.0)

        # Assert they've been properly flushed.
        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 2

        (first, second) = metrics
        assert first['metric'] == 'test.gauge.1'
        assert first['points'][0][0] == 100.0
        assert first['points'][0][1] == 22
        assert second['metric'] == 'test.gauge.2'

        # Flush again and make sure we're progressing.
        reporter.metrics = []
        dog.flush(130.0)
        assert len(reporter.metrics) == 1

        # Finally, make sure we've flushed all metrics.
        reporter.metrics = []
        dog.flush(150.0)
        assert len(reporter.metrics) == 0

    def test_set(self):
        # Create some fake metrics.
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        dog.set('test.set.1', "a string", 100.0)
        dog.set('test.set.1', frozenset(), 105.0)
        dog.set('test.set.2', 30, 115.0)
        dog.set('test.set.3', 30, 125.0)
        dog.flush(120.0)

        # Assert they've been properly flushed.
        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 2

        (first, second) = metrics
        assert first['metric'] == 'test.set.1'
        assert first['points'][0][0] == 100.0
        assert first['points'][0][1] == 2
        assert second['metric'] == 'test.set.2'
        assert second['points'][0][0] == 110.0
        assert second['points'][0][1] == 1

        # Flush again and make sure we're progressing.
        reporter.metrics = []
        dog.flush(130.0)
        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 1
        assert metrics[0]['metric'] == 'test.set.3'
        assert metrics[0]['points'][0][0] == 120.0
        assert metrics[0]['points'][0][1] == 1

        # Finally, make sure we've flushed all metrics.
        reporter.metrics = []
        dog.flush(150.0)
        assert len(reporter.metrics) == 0

    def test_counter(self):
        # Create some fake metrics.
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        dog.increment('test.counter.1', timestamp=1000.0)
        dog.increment('test.counter.1', value=2, timestamp=1005.0)
        dog.increment('test.counter.2', timestamp=1015.0)
        dog.increment('test.counter.3', timestamp=1025.0)
        dog.flush(1021.0)

        # Assert they've been properly flushed.
        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 2
        (first, second) = metrics
        assert first['metric'] == 'test.counter.1'
        assert first['points'][0][0] == 1000.0
        assert first['points'][0][1] == 0.3
        assert second['metric'] == 'test.counter.2'

        # Test decrement
        dog.increment('test.counter.1', value=10, timestamp=1000.0)
        dog.decrement('test.counter.1', value=2, timestamp=1005.0)
        reporter.metrics = []
        dog.flush(1021.0)

        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 1
        first, = metrics
        assert first['metric'] == 'test.counter.1'
        assert first['points'][0][0] == 1000.0
        assert first['points'][0][1] == 0.8
        assert second['metric'] == 'test.counter.2'

        # Flush again and make sure we're progressing.
        reporter.metrics = []
        dog.flush(1030.0)
        assert len(reporter.metrics) == 1

        # Finally, make sure we've flushed all metrics.
        reporter.metrics = []
        dog.flush(1050.0)
        assert len(reporter.metrics) == 0

    def test_distribution(self):
        # Create some fake metrics.
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        dog.distribution('test.dist.1', 20, 100.0)
        dog.distribution('test.dist.1', 22, 105.0)
        dog.distribution('test.dist.2', 30, 115.0)
        dog.distribution('test.dist.3', 30, 125.0)
        dog.flush(120.0)

        # Assert they've been properly flushed.
        dists = self.sort_metrics(reporter.distributions)
        assert len(dists) == 2

        (first, second) = dists
        assert first['metric'] == 'test.dist.1'
        assert first['points'][0][0] == 100.0
        assert first['points'][0][1] == [20, 22]
        assert second['metric'] == 'test.dist.2'

        # Flush again and make sure we're progressing.
        reporter.distributions = []
        dog.flush(130.0)
        assert len(reporter.distributions) == 1

        # Finally, make sure we've flushed all metrics.
        reporter.distributions = []
        dog.flush(150.0)
        assert len(reporter.distributions) == 0

    def test_default_host_and_device(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=1, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()
        dog.gauge('my.gauge', 1, 100.0)
        dog.flush(1000)
        metric = reporter.metrics[0]
        assert not metric['device']
        assert not metric['host']

    def test_custom_host_and_device(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=1, flush_in_thread=False, device='dev')
        reporter = dog.reporter = MemoryReporter()
        dog.gauge('my.gauge', 1, 100.0, host='host')
        dog.flush(1000)
        metric = reporter.metrics[0]
        assert metric['device'] == 'dev'
        assert metric['host'] == 'host'

    def test_tags(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Post the same metric with different tags.
        dog.gauge('gauge', 10, timestamp=100.0)
        dog.gauge('gauge', 15, timestamp=100.0, tags=['env:production', 'db'])
        dog.gauge('gauge', 20, timestamp=100.0, tags=['env:staging'])

        dog.increment('counter', timestamp=100.0)
        dog.increment('counter', timestamp=100.0, tags=['env:production', 'db'])
        dog.increment('counter', timestamp=100.0, tags=['env:staging'])

        dog.flush(200.0)

        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 6

        [c1, c2, c3, g1, g2, g3] = metrics
        assert c1['metric'] == 'counter'
        assert c2['metric'] == 'counter'
        assert c3['metric'] == 'counter'
        assert c1['tags'] is None
        assert c1['points'][0][1] == 0.1
        assert c2['tags'] == ['env:production', 'db']
        assert c2['points'][0][1] == 0.1
        assert c3['tags'] == ['env:staging']
        assert c3['points'][0][1] == 0.1

        assert g1['metric'] == 'gauge'
        assert g2['metric'] == 'gauge'
        assert g3['metric'] == 'gauge'
        assert g1['tags'] is None
        assert g1['points'][0][1] == 10
        assert g2['tags'] == ['env:production', 'db']
        assert g2['points'][0][1] == 15
        assert g3['tags'] == ['env:staging']
        assert g3['points'][0][1] == 20

    def test_constant_tags(self):
        """
        Constant tags are attached to all metrics.
        """
        dog = ThreadStats(constant_tags=["type:constant"])
        dog.start(roll_up_interval=1, flush_in_thread=False)
        dog.reporter = self.reporter

        # Post the same metric with different tags.
        dog.gauge("gauge", 10, timestamp=100.0)
        dog.gauge("gauge", 15, timestamp=100.0, tags=["env:production", 'db'])
        dog.gauge("gauge", 20, timestamp=100.0, tags=["env:staging"])

        dog.increment("counter", timestamp=100.0)
        dog.increment("counter", timestamp=100.0, tags=["env:production", 'db'])
        dog.increment("counter", timestamp=100.0, tags=["env:staging"])

        dog.flush(200.0)

        # Assertions on all metrics
        self.assertMetric(count=6)

        # Assertions on gauges
        self.assertMetric(name='gauge', value=10, tags=["type:constant"], count=1)
        self.assertMetric(name="gauge", value=15,
                          tags=["env:production", "db", "type:constant"], count=1)  # noqa
        self.assertMetric(name="gauge", value=20, tags=["env:staging", "type:constant"], count=1)

        # Assertions on counters
        self.assertMetric(name="counter", value=1, tags=["type:constant"], count=1)
        self.assertMetric(name="counter", value=1,
                          tags=["env:production", "db", "type:constant"], count=1)  # noqa
        self.assertMetric(name="counter", value=1, tags=["env:staging", "type:constant"], count=1)

        # Ensure histograms work as well.
        @dog.timed('timed', tags=['version:1'])
        def do_nothing():
            """
            A function that does nothing, but being timed.
            """
            pass

        with patch("datadog.threadstats.base.time", return_value=300):
            do_nothing()

        dog.histogram('timed', 20, timestamp=300.0, tags=['db', 'version:2'])

        self.reporter.metrics = []
        dog.flush(400.0)

        # Histograms, and related metric types, produce 8 different metrics
        self.assertMetric(tags=["version:1", "type:constant"], count=8)
        self.assertMetric(tags=["db", "version:2", "type:constant"], count=8)

    def test_metric_namespace(self):
        """
        Namespace prefixes all metric names.
        """
        # Set up ThreadStats with a namespace
        dog = ThreadStats(namespace="foo")
        dog.start(roll_up_interval=1, flush_in_thread=False)
        dog.reporter = self.reporter

        # Send a few metrics
        dog.gauge("gauge", 20, timestamp=100.0)
        dog.increment("counter", timestamp=100.0)
        dog.flush(200.0)

        # Metric names are prefixed with the namespace
        self.assertMetric(count=2)
        self.assertMetric(name="foo.gauge", count=1)
        self.assertMetric(name="foo.counter", count=1)

    def test_host(self):
        dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Post the same metric with different tags.
        dog.gauge('gauge', 12, timestamp=100.0, host='')  # unset the host
        dog.gauge('gauge', 10, timestamp=100.0)
        dog.gauge('gauge', 15, timestamp=100.0, host='test')
        dog.gauge('gauge', 15, timestamp=100.0, host='test')

        dog.increment('counter', timestamp=100.0)
        dog.increment('counter', timestamp=100.0)
        dog.increment('counter', timestamp=100.0, host='test')
        dog.increment('counter', timestamp=100.0, host='test', tags=['tag'])
        dog.increment('counter', timestamp=100.0, host='test', tags=['tag'])

        dog.flush(200.0)

        metrics = self.sort_metrics(reporter.metrics)
        assert len(metrics) == 6

        [c1, c2, c3, g1, g2, g3] = metrics
        assert c1['metric'] == 'counter'
        assert c2['metric'] == 'counter'
        assert c3['metric'] == 'counter'
        assert c1['host'] is None
        assert c1['tags'] is None
        assert c1['points'][0][1] == 0.2
        assert c2['host'] == 'test'
        assert c2['tags'] is None
        assert c2['points'][0][1] == 0.1
        assert c3['host'] == 'test'
        assert c3['tags'] == ['tag']
        assert c3['points'][0][1] == 0.2

        assert g1['metric'] == 'gauge'
        assert g2['metric'] == 'gauge'
        assert g3['metric'] == 'gauge'
        assert g1['host'] is None
        assert g1['points'][0][1] == 10
        assert g2['host'] == ''
        assert g2['points'][0][1] == 12
        assert g3['host'] == 'test'
        assert g3['points'][0][1] == 15

        # Ensure histograms work as well.
        @dog.timed('timed', host='test')
        def test():
            pass
        test()
        dog.histogram('timed', 20, timestamp=300.0, host='test')
        reporter.metrics = []
        dog.flush(400)
        for metric in reporter.metrics:
            assert metric['host'] == 'test'

    def test_disabled_mode(self):
        dog = ThreadStats()
        dog.start(disabled=True, flush_interval=1, roll_up_interval=1)
        reporter = dog.reporter = MemoryReporter()
        dog.gauge('testing', 1, timestamp=1000)
        dog.gauge('testing', 2, timestamp=1000)
        dog.flush(2000.0)
        assert not reporter.metrics

    def test_stop(self):
        dog = ThreadStats()
        dog.start(flush_interval=1, roll_up_interval=1)
        dog.reporter = MemoryReporter()
        for i in range(10):
            dog.gauge('metric', i)
        time.sleep(2)
        flush_count = dog.flush_count
        assert flush_count
        dog.stop()
        for i in range(10):
            dog.gauge('metric', i)
        time.sleep(2)
        for i in range(10):
            dog.gauge('metric', i)
        time.sleep(2)
        assert dog.flush_count in [flush_count, flush_count + 1]

    def test_tags_from_environment(self):
        test_tags = ['country:china', 'age:45', 'blue']
        with preserve_environment_variable('DATADOG_TAGS'):
            os.environ['DATADOG_TAGS'] = ','.join(test_tags)
            dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Add two events
        event1_title = "Event 1 title"
        event2_title = "Event 1 title"
        event1_text = "Event 1 text"
        event2_text = "Event 2 text"
        dog.event(event1_title, event1_text)
        dog.event(event2_title, event2_text)

        # Flush and test
        dog.flush()
        event1, event2 = reporter.events
        assert event1['title'] == event1_title
        assert event1['text'] == event1_text
        assert event1['tags'] == test_tags
        assert event2['title'] == event2_title
        assert event2['text'] == event2_text
        assert event2['text'] == event2_text
        assert event2['tags'] == test_tags

        # Test more parameters
        reporter.events = []
        event1_priority = "low"
        event1_date_happened = 1375296969
        event1_tag = "Event 2 tag"
        dog.event(event1_title, event1_text, priority=event1_priority,
                  date_happened=event1_date_happened, tags=[event1_tag])

        # Flush and test
        dog.flush()
        event, = reporter.events
        assert event['title'] == event1_title
        assert event['text'] == event1_text
        assert event['priority'] == event1_priority
        assert event['date_happened'] == event1_date_happened
        assert event['tags'] == [event1_tag] + test_tags
        dog.start(flush_interval=1, roll_up_interval=1)

    def test_tags_from_environment_and_constant(self):
        test_tags = ['country:china', 'age:45', 'blue']
        constant_tags = ['country:canada', 'red']
        with preserve_environment_variable('DATADOG_TAGS'):
            os.environ['DATADOG_TAGS'] = ','.join(test_tags)
            dog = ThreadStats(constant_tags=constant_tags)
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Add two events
        event1_title = "Event 1 title"
        event2_title = "Event 1 title"
        event1_text = "Event 1 text"
        event2_text = "Event 2 text"
        dog.event(event1_title, event1_text)
        dog.event(event2_title, event2_text)

        # Flush and test
        dog.flush()
        event1, event2 = reporter.events
        assert event1['title'] == event1_title
        assert event1['text'] == event1_text
        assert event1['tags'] == constant_tags + test_tags
        assert event2['title'] == event2_title
        assert event2['text'] == event2_text
        assert event2['text'] == event2_text
        assert event2['tags'] == constant_tags + test_tags

        # Test more parameters
        reporter.events = []
        event1_priority = "low"
        event1_date_happened = 1375296969
        event1_tag = "Event 2 tag"
        dog.event(event1_title, event1_text, priority=event1_priority,
                  date_happened=event1_date_happened, tags=[event1_tag])

        # Flush and test
        dog.flush()
        event, = reporter.events
        assert event['title'] == event1_title
        assert event['text'] == event1_text
        assert event['priority'] == event1_priority
        assert event['date_happened'] == event1_date_happened
        assert event['tags'] == [event1_tag] + constant_tags + test_tags
        dog.start(flush_interval=1, roll_up_interval=1)

    def test_tags_from_environment_env_service_version(self):
        test_tags = set(['env:staging', 'service:food', 'version:1.2.3'])
        with EnvVars(
            env_vars={
                "DD_ENV": "staging",
                "DD_VERSION": "1.2.3",
                "DD_SERVICE": "food",
            }
        ):
            dog = ThreadStats()
        dog.start(roll_up_interval=10, flush_in_thread=False)
        reporter = dog.reporter = MemoryReporter()

        # Add two events
        event1_title = "Event 1 title"
        event1_text = "Event 1 text"
        dog.event(event1_title, event1_text)

        # Flush and test
        dog.flush()
        [event1] = reporter.events
        assert event1['title'] == event1_title
        assert event1['text'] == event1_text
        assert set(event1['tags']) == test_tags

    def test_metric_type(self):
        """
        Checks the submitted metric's metric type.
        """
        # Set up ThreadStats with a namespace
        dog = ThreadStats(namespace="foo")
        dog.start(roll_up_interval=1, flush_in_thread=False)
        reporter = dog.reporter = self.reporter

        # Send a few metrics
        dog.gauge("gauge", 20, timestamp=100.0)
        dog.increment("counter", timestamp=100.0)
        dog.histogram('histogram.1', 20, 100.0)
        dog.flush(200.0)

        (first, second, p75, p85, p95, p99, avg, cnt, max_, min_) = self.sort_metrics(reporter.metrics)

        # Assert Metric type
        assert first['type'] == 'rate'
        assert second['type'] == 'gauge'
        assert p75['type'] == 'gauge'
        assert p85['type'] == 'gauge'
        assert p95['type'] == 'gauge'
        assert p99['type'] == 'gauge'
        assert avg['type'] == 'gauge'
        assert cnt['type'] == 'rate'
        assert max_['type'] == 'gauge'
        assert min_['type'] == 'gauge'

    # Test lambda_wrapper (uses ThreadStats under the hood)
    def test_basic_lambda_decorator(self):

        @datadog_lambda_wrapper
        def basic_wrapped_function():
            lambda_metric("lambda.somemetric", 100)

        _get_lambda_stats().reporter = self.reporter
        basic_wrapped_function()

        assert _get_lambda_stats().reporter.dist_flush_counter == 1
        dists = self.sort_metrics(_get_lambda_stats().reporter.distributions)
        assert len(dists) == 1

    def test_embedded_lambda_decorator(self):
        """
        Test that the lambda decorator flushes metrics correctly and only once
        """

        @datadog_lambda_wrapper
        def wrapped_function_1():
            lambda_metric("lambda.dist.1", 10)

        @datadog_lambda_wrapper
        def wrapped_function_2():
            wrapped_function_1()
            lambda_metric("lambda.dist.2", 30)

        _get_lambda_stats().reporter = self.reporter
        wrapped_function_2()
        assert _get_lambda_stats().reporter.dist_flush_counter == 1

        dists = self.sort_metrics(_get_lambda_stats().reporter.distributions)
        assert len(dists) == 2



# coding: utf8
# Unless explicitly stated otherwise all files in this repository are licensed
# under the BSD-3-Clause License. This product includes software developed at
# Datadog (https://www.datadoghq.com/).

# Copyright 2015-Present Datadog, Inc

# stdlib
import cProfile
import io
import logging
import os
import pstats
import random
import sys
import threading
import timeit
import unittest
import warnings

try:
    import queue
except ImportError:
    import Queue as queue

# datadog
from datadog.dogstatsd.base import DogStatsd
from datadog.util.compat import is_p3k

# test utils
from tests.util.fake_statsd_server import FakeServer
from tests.util.system_info_observer import SysInfoObserver


# StatsdSender is a static helper for sending mock metrics to statsd via a simple API
# pylint: disable=too-few-public-methods,useless-object-inheritance
class StatsdSender(object):
    EXTRA_TAGS = ["bar = barval", "baz = bazval"]
    STATIC_TIMING_SET = set(range(100))

    # Enums are not part of 2.7 built-ins
    METRICS_TYPE = [
        "decrement",
        "distribution",
        "gauge",
        "histogram",
        "increment",
        "set",
        "timing",
    ]

    @staticmethod
    def send(metric, statsd_instance, value):
        getattr(StatsdSender, "_submit_{}".format(StatsdSender.METRICS_TYPE[metric]))(
            statsd_instance, threading.current_thread().name, value
        )

    @staticmethod
    def _submit_decrement(statsd_instance, metric_prefix, _):
        statsd_instance.decrement(
            "{}.counter".format(metric_prefix), tags=StatsdSender.EXTRA_TAGS
        )

    @staticmethod
    def _submit_distribution(statsd_instance, metric_prefix, value):
        statsd_instance.distribution(
            "{}.distribution".format(metric_prefix), value, tags=StatsdSender.EXTRA_TAGS
        )

    @staticmethod
    def _submit_gauge(statsd_instance, metric_prefix, value):
        statsd_instance.gauge(
            "{}.gauge".format(metric_prefix), value, tags=StatsdSender.EXTRA_TAGS
        )

    @staticmethod
    def _submit_histogram(statsd_instance, metric_prefix, value):
        statsd_instance.histogram(
            "{}.histogram".format(metric_prefix), value, tags=StatsdSender.EXTRA_TAGS
        )

    @staticmethod
    def _submit_increment(statsd_instance, metric_prefix, _):
        statsd_instance.increment(
            "{}.counter".format(metric_prefix), tags=StatsdSender.EXTRA_TAGS
        )

    @staticmethod
    def _submit_set(statsd_instance, metric_prefix, value):
        statsd_instance.set(
            "{}.set".format(metric_prefix), value, tags=StatsdSender.EXTRA_TAGS
        )

    @staticmethod
    def _submit_timing(statsd_instance, metric_prefix, _):
        statsd_instance.timing(
            "{}.set".format(metric_prefix),
            StatsdSender.STATIC_TIMING_SET,
            tags=StatsdSender.EXTRA_TAGS,
        )


class TestDogStatsdThroughput(unittest.TestCase):
    """
    DogStatsd throughput tests.
    """

    DEFAULT_NUM_DATAPOINTS = 50000
    DEFAULT_NUM_THREADS = 1
    DEFAULT_NUM_RUNS = 5
    DEFAULT_TRANSPORT = "udp"

    RUN_MESSAGE = (
        "Run #{:2d}/{:2d}: {:.4f}s (latency: {:.2f}μs, cpu: {:.4f},"
        + " mem.rss_diff: {:.0f}kb, recv: {:.2f}%)"
    )

    def setUp(self):
        # Parse the benchmark parameters and use sensible defaults for values
        # that are not configured
        self.num_datapoints = int(
            os.getenv("BENCHMARK_NUM_DATAPOINTS", str(self.DEFAULT_NUM_DATAPOINTS))
        )
        self.num_threads = int(
            os.getenv("BENCHMARK_NUM_THREADS", str(self.DEFAULT_NUM_THREADS))
        )
        self.num_runs = int(os.getenv("BENCHMARK_NUM_RUNS", str(self.DEFAULT_NUM_RUNS)))
        self.profiling_enabled = os.getenv("BENCHMARK_PROFILING", "false") in ["1", "true", "True", "Y", "yes", "Yes"]
        self.transport = os.getenv(
            "BENCHMARK_TRANSPORT", str(self.DEFAULT_TRANSPORT)
        ).upper()

        # We do want to see any problems if they occur in the statsd library
        logger = logging.getLogger()
        logger.level = logging.DEBUG
        logger.addHandler(logging.StreamHandler(sys.stdout))

        # Ensure that warnings don't print the stack trace
        def one_line_warning(message, category, filename, lineno, *_):
            return "%s:%s: %s: %s" % (filename, lineno, category.__name__, message)

        warnings.formatwarning = one_line_warning

        # Add a newline so that we don't get clobbered by the test output
        print("")

    # pylint: disable=too-many-locals
    def test_statsd_performance(self):
        print(
            "Starting: {} run(s), {} thread(s), {} points/thread via {} (profiling: {}) on Python{}.{} ...".format(
                self.num_runs,
                self.num_threads,
                self.num_datapoints,
                self.transport,
                str(self.profiling_enabled).lower(),
                sys.version_info[0],
                sys.version_info[1],
            )
        )

        # We want a stable random sequence so that parallel runs
        # are consistent and repeatable
        random.seed(1234)

        # Pre-calculate a random order of metric types for each thread
        metrics_order = []
        for _ in range(self.num_threads):
            thread_metrics_order = []
            for _ in range(self.num_datapoints):
                thread_metrics_order.append(
                    random.randrange(len(StatsdSender.METRICS_TYPE))
                )

            metrics_order.append(thread_metrics_order)

        run_cpu_stats = []
        run_durations = []
        run_latencies = []
        run_memory_stats = []
        received_packet_pcts = []

        for run_idx in range(self.num_runs):
            (
                duration,
                total_latency,
                sys_stats,
                received_packet_pct,
            ) = self._execute_test_run(
                FakeServer(transport=self.transport),
                metrics_order,
                self.num_threads,
                self.num_datapoints,
            )
            avg_latency_secs = total_latency / (self.num_threads * self.num_datapoints)
            avg_latency = avg_latency_secs * 1000000
            print(
                self.RUN_MESSAGE.format(
                    run_idx + 1,
                    self.num_runs,
                    duration,
                    avg_latency,
                    sys_stats["cpu.user"] + sys_stats["cpu.system"],
                    sys_stats["mem.rss_diff_kb"],
                    received_packet_pct,
                )
            )

            run_durations.append(duration)
            run_cpu_stats.append(sys_stats["cpu.user"] + sys_stats["cpu.system"])
            run_memory_stats.append(sys_stats["mem.rss_diff_kb"])
            run_latencies.append(float(avg_latency))
            received_packet_pcts.append(received_packet_pct)

        result_msg = "\nTotal for {} run(s), {} thread(s), {} points/thread via {} on Python{}.{}:\n"
        result_msg += "\tDuration:\t\t{:.4f}s\n"
        result_msg += "\tLatency:\t\t{:.2f}μs\n"
        result_msg += "\tCPU:\t\t\t{:.4f}\n"
        result_msg += "\tMemory (rss) diff:\t{:.0f}kb\n"
        result_msg += "\tReceived packets:\t{:.2f}%"
        print(
            result_msg.format(
                self.num_runs,
                self.num_threads,
                self.num_datapoints,
                self.transport,
                sys.version_info[0],
                sys.version_info[1],
                sum(run_durations) / len(run_durations),
                sum(run_latencies) / len(run_latencies),
                sum(run_cpu_stats) / len(run_cpu_stats),
                sum(run_memory_stats) / len(run_memory_stats),
                sum(received_packet_pcts) / len(received_packet_pcts),
            )
        )

    # pylint: disable=too-many-locals,no-self-use
    def _execute_test_run(self, server, metrics_order, num_threads, num_datapoints):
        # Setup all the threads and get them in a waiting state
        threads = []
        start_signal = threading.Event()

        latency_results = queue.Queue()
        observer = SysInfoObserver()

        with server:
            # Create a DogStatsd client with a mocked socket
            statsd_instance = DogStatsd(
                constant_tags=["foo = {}".format(random.random())],
                host="localhost",
                port=server.port,
                socket_path=server.socket_path,
            )

            for thread_idx in range(num_threads):
                thread = threading.Thread(
                    name="test_statsd_throughput_thread_{}".format(thread_idx),
                    target=TestDogStatsdThroughput._thread_runner,
                    args=(
                        statsd_instance,
                        start_signal,
                        metrics_order[thread_idx],
                        latency_results,
                        self.profiling_enabled,
                    ),
                )
                thread.daemon = True
                threads.append(thread)
                thread.start()

            # `timeit.timeit` is not easily usable here since we need to pass in state
            # and Python 2 version of `timeit()` does not accept the `global` keyword.
            start_time = timeit.default_timer()

            # Let the thread know that it can start sending metrics
            start_signal.set()

            # Observe system utilization while we wait for the threads to exit
            with observer:
                for thread in threads:
                    thread.join()

            total_latency = 0.0
            for thread in threads:
                if latency_results.empty():
                    warnings.warn("One or more threads did not report their results!")
                    continue

                total_latency += latency_results.get()

            duration = timeit.default_timer() - start_time

        # Sanity checks: Verify that metric transfer expectations are correct
        expected_metrics = num_threads * num_datapoints

        # Verify that dropped metric count is matching our statsd expectations. This
        # type of inconsistency should never happen.
        if (
            expected_metrics - server.metrics_captured
            != statsd_instance.packets_dropped
        ):
            error_msg = (
                "WARN: Statsd dropped packet count ({}) did not match the server "
            )
            error_msg += "missing received packet count expectation ({})!\n"
            warnings.warn(
                error_msg.format(
                    statsd_instance.packets_dropped,
                    expected_metrics - server.metrics_captured,
                )
            )

        # Verify that received metric count is matching our metric totals expectations. Note
        # that in some scenarios, some data is expected to be dropped.
        if server.metrics_captured != expected_metrics:
            error_msg = "WARN: Received metrics count ({}) did not match the sent "
            error_msg += "metrics count ({})!\n"
            warnings.warn(error_msg.format(server.metrics_captured, expected_metrics))

        received_packet_pct = server.metrics_captured * 100.0 / expected_metrics

        return (duration, total_latency, observer.stats, received_packet_pct)

    @staticmethod
    def _thread_runner(
        statsd_instance,
        start_event,
        thread_metrics_order,
        latency_results,
        profiling_enabled,
    ):
        # We wait for a global signal to start running our events
        start_event.wait(5)

        if profiling_enabled:
            profiler = cProfile.Profile()
            profiler.enable()

        duration = 0.0
        for metric_idx, metric in enumerate(thread_metrics_order):
            start_time = timeit.default_timer()

            StatsdSender.send(metric, statsd_instance, metric_idx)

            duration += timeit.default_timer() - start_time

        if hasattr(statsd_instance, 'flush'):
            statsd_instance.flush()

        latency_results.put(duration)

        if profiling_enabled:
            TestDogStatsdThroughput.print_profiling_stats(profiler)


    @staticmethod
    def print_profiling_stats(profiler, sort_by='cumulative'):
        """
        Prints profiling results for the thread that finishes its run. Options for
        sorting include 'tottime', 'pcalls', 'ncalls', 'cumulative', etc but you can
        check https://github.com/python/cpython/blob/3.9/Lib/pstats.py#L37-L45 for
        other options.
        """

        profiler.disable()

        if is_p3k():
            output_stream = io.StringIO()
        else:
            output_stream = io.BytesIO()

        profiling_stats = pstats.Stats(
                profiler,
                stream=output_stream,
        ).sort_stats(sort_by)

        profiling_stats.print_stats()
        print(output_stream.getvalue())



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
import os

API_KEY = os.environ.get("DD_TEST_CLIENT_API_KEY", "a" * 32)
APP_KEY = os.environ.get("DD_TEST_CLIENT_APP_KEY", "a" * 40)
API_HOST = os.environ.get("DATADOG_HOST")
MONITOR_REFERENCED_IN_SLO_MESSAGE = "monitor {} is referenced in slos: {}"
TEST_USER = os.environ.get("DD_TEST_CLIENT_USER")



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc

from itertools import product

import pytest

TEST_ACCOUNT_ID_1 = "123456789101"
TEST_ACCOUNT_ID_2 = "123456789102"
TEST_ACCOUNT_ID_3 = "123456789103"
TEST_ACCOUNT_ID_4 = "123456789104"
TEST_ROLE_NAME = "DatadogApiTestRole"
TEST_ROLE_NAME_2 = "DatadogApiTestRolo"

ROLE_NAMES = [TEST_ROLE_NAME, TEST_ROLE_NAME_2]
ACCOUNT_IDS = [TEST_ACCOUNT_ID_1, TEST_ACCOUNT_ID_2, TEST_ACCOUNT_ID_3, TEST_ACCOUNT_ID_4]
AVAILABLE_NAMESPACES = 76

class TestAwsIntegration:

    @pytest.fixture(autouse=True)  # TODO , scope="class"
    def aws_integration(self, dog):
        """Remove pending AWS Integrations."""
        yield
        for account_id, role_name in product(ACCOUNT_IDS, ROLE_NAMES):
            dog.AwsIntegration.delete(account_id=account_id, role_name=role_name)

    def test_create(self, dog):
        output = dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_3,
            role_name=TEST_ROLE_NAME,
            host_tags=["api:test"],
            filter_tags=["filter:test"],
            account_specific_namespace_rules={'auto_scaling': False, 'opsworks': False}
        )
        assert "external_id" in output

    def test_list(self, dog):
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_1,
            role_name=TEST_ROLE_NAME
        )
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME
        )

        output = dog.AwsIntegration.list()
        assert "accounts" in output
        assert len(output['accounts']) >= 2
        expected_fields = [
            'errors',
            'filter_tags',
            'host_tags',
            'account_specific_namespace_rules',
        ]
        assert all(k in output['accounts'][0].keys() for k in expected_fields)

    def test_delete(self, dog):
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_1,
            role_name=TEST_ROLE_NAME
        )
        output = dog.AwsIntegration.delete(account_id=TEST_ACCOUNT_ID_1, role_name=TEST_ROLE_NAME)
        assert output == {}

    def test_generate_new_external_id(self, dog):
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME
        )
        output = dog.AwsIntegration.generate_new_external_id(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME
        )

        assert "external_id" in output

    def test_list_namespace_rules(self, dog):
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME
        )
        output = dog.AwsIntegration.list_namespace_rules(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME
        )
        assert len(output) >= AVAILABLE_NAMESPACES

    def test_update(self, dog):
        dog.AwsIntegration.create(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME
        )

        dog.AwsIntegration.update(
            account_id=TEST_ACCOUNT_ID_2,
            role_name=TEST_ROLE_NAME,
            new_account_id=TEST_ACCOUNT_ID_4,
            host_tags=["api:test2"],
            new_role_name=TEST_ROLE_NAME_2,
            excluded_regions=["us-east-1","us-west-1"]
        )

        output = dog.AwsIntegration.list()
        tests_pass = False
        for i in output['accounts']:
            assert "excluded_regions" in i.keys()
            if i.get('account_id') == TEST_ACCOUNT_ID_4 and i.get('role_name') == TEST_ROLE_NAME_2 and i.get('excluded_regions') == ["us-east-1","us-west-1"]:
                tests_pass = True
        assert tests_pass
